#!/usr/bin/env python3
#
# Tests the WCP format module.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from __future__ import absolute_import, division
from __future__ import print_function, unicode_literals

import os
import unittest
import numpy as np

import myokit.formats.wcp as wcp

from myokit.tests import DIR_FORMATS


class WcpTest(unittest.TestCase):
    """
    Tests the WCP format module.
    """

    def test_data_file(self):
        # Test basic wcp file reading.

        # Load old file from Maastricht
        fname = 'wcp-file.wcp'
        path = os.path.join(DIR_FORMATS, fname)
        w = wcp.WcpFile(path)

        self.assertEqual(w.channels(), 2)
        self.assertEqual(w.channel_names(), ['Im', 'Vm'])
        self.assertEqual(w.filename(), fname)
        self.assertEqual(w.path(), path)
        self.assertEqual(w.records(), 11)
        #print(w.sampling_interval())
        self.assertEqual(len(w.times()), 256)
        self.assertEqual(len(w.values(0, 0)), 256)

    def test_data_log_conversion(self):
        # Test conversion to a data log.

        w = wcp.WcpFile(os.path.join(DIR_FORMATS, 'wcp-file.wcp'))
        d = w.myokit_log()
        keys = ['time']
        for cn in w.channel_names():
            for i in range(w.records()):
                keys.append(str(i) + '.' + cn)
        self.assertEqual(set(d.keys()), set(keys))
        self.assertEqual(len(d.time()), 256)
        self.assertTrue(np.all(d.time() == w.times()))

    def test_plot_method(self):
        # Test the plot method.

        # Select matplotlib backend that doesn't require a screen
        import matplotlib
        matplotlib.use('template')

        # Load and create plots
        w = wcp.WcpFile(os.path.join(DIR_FORMATS, 'wcp-file.wcp'))
        w.plot()


if __name__ == '__main__':
    unittest.main()



#
# CVODE-driven single cell simulation
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from __future__ import absolute_import, division
from __future__ import print_function, unicode_literals

import os
import myokit
import platform

# Location of C template
SOURCE_FILE = 'cvodesim.c'


class Simulation(myokit.CModule):
    """
    Runs single cell simulations using the CVODE solver (see [1]); CVODE uses
    an implicit multi-step method to achieve high accuracy and stability with
    adaptive step sizes.

    The model passed to the simulation is cloned and stored internally, so
    changes to the original model object will not affect the simulation. A
    protocol can be passed in as ``protocol`` or set later using
    :meth:`set_protocol`.

    Simulations maintain an internal state consisting of

    - the current simulation time
    - the current state
    - the default state

    When a simulation is created, the simulation time is set to 0 and both the
    current and the default state are copied from the model.
    After each call to :meth:`Simulation.run` the time variable and current
    state are updated, so that each successive call to run continues where the
    previous simulation left off. A :meth:`reset` method is provided that will
    set the time back to 0 and revert the current state to the default state.
    To change the time or state manually, use :meth:`set_time` and
    :meth:`set_state`.

    A pre-pacing method :meth:`pre` is provided that doesn't affect the
    simulation time but will update the current *and the default state*. This
    allows you to pre-pace, run a simulation, reset to the pre-paced state, run
    another simulation etc.

    To get action potential duration (APD) measurements, the simulation can be
    run with threshold crossing detection. To enable this, the membrane
    potential variable *must* be specified when the simulation is created using
    the ``apd_var`` argument. This can be either a variable object or a string
    containing the variable's fully qualified name. When running a simulation a
    threshold value can be passed in. In addition to the usual simulation log
    the run method will then return a list of all times at which ``apd_var``
    crossed the threshold. *Please note this is an APD calculated as the time
    between the crossing of a fixed threshold, it does not calculate dynamic
    thresholds like "90% of max(V) - min(V)".*

    The simulation provides four inputs a model variable can be bound to:

    ``time``
        This input provides the simulation time.
    ``pace``
        This input provides the current value of the pacing variable. This is
        determined using the protocol passed into the Simulation.
    ``evaluations``
        This input provides the number of rhs evaluations used at each point in
        time and can be used to gain some insight into the solver's behaviour.
    ``realtime``
        This input provides the elapsed system time at each logged point.

    No variable labels are required for this simulation type.

    [1] SUNDIALS: Suite of nonlinear and differential/algebraic equation
    solvers. Hindmarsh, Brown, Woodward, et al. (2005) ACM Transactions on
    Mathematical Software.

    """
    _index = 0  # Simulation id

    def __init__(self, model, protocol=None, apd_var=None):
        super(Simulation, self).__init__()

        # Require a valid model
        if not model.is_valid():
            model.validate()
        model = model.clone()
        self._model = model

        # Set protocol
        self._protocol = None
        self._fixed_form_protocol = None
        self.set_protocol(protocol)

        # Check potential and threshold values
        if apd_var is None:
            self._apd_var = None
        else:
            if isinstance(apd_var, myokit.Variable):
                apd_var = apd_var.qname()
            self._apd_var = self._model.get(apd_var)
            if not self._apd_var.is_state():
                raise ValueError('The `apd_var` must be a state variable.')

        # Get state and default state from model
        self._state = self._model.state()
        self._default_state = list(self._state)

        # Last state reached before error
        self._error_state = None

        # Starting time
        self._time = 0

        # Unique simulation id
        Simulation._index += 1
        module_name = 'myokit_leg_sim_' + str(Simulation._index)
        module_name += '_' + str(myokit.pid_hash())

        # Arguments
        args = {
            'module_name': module_name,
            'model': self._model,
            'potential': self._apd_var,
        }
        fname = os.path.join(myokit.DIR_CFUNC, SOURCE_FILE)

        # Define libraries
        libs = [
            'sundials_cvode',
            'sundials_nvecserial',
        ]
        if platform.system() != 'Windows':  # pragma: no windows cover
            libs.append('m')

        # Define library paths
        # Note: Sundials path on windows already includes local binaries
        libd = list(myokit.SUNDIALS_LIB)
        incd = list(myokit.SUNDIALS_INC)
        incd.append(myokit.DIR_CFUNC)

        # Create extension
        self._sim = self._compile(module_name, fname, args, libs, libd, incd)

        # Set default tolerance values
        self._tolerance = None
        self.set_tolerance()

        # Set default min and max step size
        self._dtmax = self._dtmin = None

    def default_state(self):
        """
        Returns the default state.
        """
        return list(self._default_state)

    def last_state(self):
        """
        If the last simulation resulted in an error, this will return the last
        state reached during that simulation. In all other cases, this method
        will return ``None``.
        """
        return list(self._error_state) if self._error_state else None

    def eval_derivatives(self, y=None):
        """
        Evaluates and returns the state derivatives.

        The state to evaluate for can be given as ``y``. If no state is given
        the current simulation state is used.
        """
        if y is None:
            y = list(self._state)
        else:
            y = self._model.map_to_state(y)
        dy = list(self._state)
        self._sim.eval_derivatives(y, dy, 0, 0)
        return dy

    def last_number_of_evaluations(self):
        """
        Returns the number of rhs evaluations performed by the solver during
        the last simulation.
        """
        return self._sim.number_of_evaluations()

    def last_number_of_steps(self):
        """
        Returns the number of steps taken by the solver during the last
        simulation.
        """
        return self._sim.number_of_steps()

    def pre(self, duration, progress=None, msg='Pre-pacing Simulation'):
        """
        This method can be used to perform an unlogged simulation, typically to
        pre-pace to a (semi-)stable orbit.

        After running this method

        - The simulation time is **not** affected
        - The current state and the default state are updated to the final
          state reached in the simulation.

        Calls to :meth:`reset` after using :meth:`pre` will set the current
        state to this new default state.

        To obtain feedback on the simulation progress, an object implementing
        the :class:`myokit.ProgressReporter` interface can be passed in.
        passed in as ``progress``. An optional description of the current
        simulation to use in the ProgressReporter can be passed in as `msg`.
        """
        duration = float(duration)
        self._run(duration, myokit.LOG_NONE, None, None, None, progress, msg)
        self._default_state = self._state

    def __reduce__(self):
        """
        Pickles this Simulation.

        See: https://docs.python.org/3/library/pickle.html#object.__reduce__
        """
        apd_var = None if self._apd_var is None else self._apd_var.qname()
        return (
            self.__class__,
            (self._model, self._protocol, apd_var),
            (
                self._time,
                self._state,
                self._default_state,
                self._fixed_form_protocol,
                self._tolerance,
                self._dtmin,
                self._dtmax,
            ),
        )

    def reset(self):
        """
        Resets the simulation:

        - The time variable is set to 0
        - The state is set to the default state

        """
        self._time = 0
        self._state = list(self._default_state)

    def run(
            self, duration, log=None, log_interval=None, log_times=None,
            apd_threshold=None, progress=None, msg='Running simulation'):
        """
        Runs a simulation and returns the logged results. Running a simulation
        has the following effects:

        - The internal state is updated to the last state in the simulation.
        - The simulation's time variable is updated to reflect the time
            elapsed during the simulation.

        The number of time units to simulate can be set with ``duration``.

        The method returns a :class:`myokit.DataLog` dictionary that maps
        variable names to lists of logged values. The variables to log can be
        indicated using the ``log`` argument. There are several options for its
        value:

        - ``None`` (default), to log all states.
        - An integer flag or a combination of flags. Options:
          ``myokit.LOG_NONE``, ``myokit.LOG_STATE``, ``myokit.LOG_BOUND``,
          ``myokit.LOG_INTER``, ``myokit.LOG_DERIV`` or ``myokit.LOG_ALL``.
        - A sequence of variable names. To log derivatives, use
          "dot(membrane.V)".
        - A :class:`myokit.DataLog` object. In this case, the new data
          will be appended to the existing log.

        For detailed information about the ``log`` argument, see the function
        :meth:`myokit.prepare_log`.

        By default, every step the solver takes is logged. This is usually
        advantageous, since more points are added exactly at the times the
        system gets more interesting. However, if equidistant points are
        required a ``log_interval`` can be set. Alternatively, the
        ``log_times`` argument can be used to specify logging times directly.

        To obtain accurate measurements of the action potential (AP) duration,
        the argument ``apd_threshold`` can be set to a fixed threshold level
        used to define the AP. This functionality is only available for
        simulations created with a valid ``apd_var`` argument. If apd
        measurements are enabled, the value returned by this method has the
        form ``(log, apds)``.

        To obtain feedback on the simulation progress, an object implementing
        the :class:`myokit.ProgressReporter` interface can be passed in.
        passed in as ``progress``. An optional description of the current
        simulation to use in the ProgressReporter can be passed in as ``msg``.

        The ``duration`` argument cannot be negative, and special care needs to
        be taken when very small (positive) values are used. If ``duration`` is
        zero or so small that
        ``simulation.time() + duration == simulation.time()``, then the method
        returns without updating the internal states or time. However, for
        some extremely short durations (approx ``2 epsilon * time``), the
        simulation will try to run but the underlying CVODE engine may return a
        ``CV_TOO_CLOSE`` error, causing a :class:`myokit.SimulationError` to be
        raised.
        """
        duration = float(duration)
        output = self._run(
            duration, log, log_interval, log_times, apd_threshold, progress,
            msg)
        self._time += duration
        return output

    def _run(
            self, duration, log, log_interval, log_times, apd_threshold,
            progress, msg):

        # Reset error state
        self._error_state = None

        # Simulation times
        if duration < 0:
            raise ValueError('Simulation time can\'t be negative.')
        tmin = self._time
        tmax = tmin + duration

        # Parse log argument
        log = myokit.prepare_log(log, self._model, if_empty=myokit.LOG_ALL)

        # Logging period (None or 0 = disabled)
        log_interval = 0 if log_interval is None else float(log_interval)
        if log_interval < 0:
            log_interval = 0

        # Logging points (None or empty list = disabled)
        if log_times is not None:
            log_times = [float(x) for x in log_times]
            if len(log_times) == 0:
                log_times = None
            else:
                # Allow duplicates, but always non-decreasing!
                import numpy as np
                x = np.asarray(log_times)
                if np.any(x[1:] < x[:-1]):
                    raise ValueError(
                        'Values in log_times must be non-decreasing.')
                del(x, np)
        if log_times is not None and log_interval > 0:
            raise ValueError(
                'The arguments log_times and log_interval cannot be used'
                ' simultaneously.')

        # Threshold for APD measurement
        root_list = None
        root_threshold = 0
        if apd_threshold is not None:
            if self._apd_var is None:
                raise ValueError(
                    'Threshold given but Simulation object was'
                    ' created without apd_var argument.')
            else:
                root_list = []
                root_threshold = float(apd_threshold)

        # Get progress indication function (if any)
        if progress is None:
            progress = myokit._Simulation_progress
        if progress:
            if not isinstance(progress, myokit.ProgressReporter):
                raise ValueError(
                    'The argument "progress" must be either a'
                    ' subclass of myokit.ProgressReporter or None.')

        # Determine benchmarking mode, create time() function if needed
        if self._model.binding('realtime') is not None:
            import timeit
            bench = timeit.default_timer
        else:
            bench = None

        # Run simulation
        # The simulation is run only if (tmin + duration > tmin). This is a
        # stronger check than (duration == 0), which will return true even for
        # very short durations (and will cause zero iterations of the
        # "while (t < tmax)" loop below).
        istate = list(self._state)
        if tmin + duration > tmin:

            # Lists to return state in
            rstate = list(istate)
            rbound = [0, 0, 0, 0]  # time, pace, realtime, evaluations

            # Initialize
            self._sim.sim_init(
                tmin,
                tmax,
                istate,
                rstate,
                rbound,
                self._protocol,
                self._fixed_form_protocol,
                log,
                log_interval,
                log_times,
                root_list,
                root_threshold,
                bench,
            )
            t = tmin

            try:
                if progress:
                    # Loop with feedback
                    with progress.job(msg):
                        r = 1.0 / duration if duration != 0 else 1
                        while t < tmax:
                            t = self._sim.sim_step()
                            if not progress.update(min((t - tmin) * r, 1)):
                                raise myokit.SimulationCancelledError()
                else:
                    # Loop without feedback
                    while t < tmax:
                        t = self._sim.sim_step()
            except ArithmeticError as e:
                # Some CVODE errors are set to raise an ArithmeticError,
                # which users may be able to debug.
                self._error_state = list(rstate)
                txt = ['A numerical error occurred during simulation at'
                       ' t = ' + str(t) + '.', 'Last reached state: ']
                txt.extend(['  ' + x for x in
                            self._model.format_state(rstate).splitlines()])
                txt.append('Inputs for binding: ')
                txt.append('  time        = ' + myokit.float.str(rbound[0]))
                txt.append('  pace        = ' + myokit.float.str(rbound[1]))
                txt.append('  realtime    = ' + myokit.float.str(rbound[2]))
                txt.append('  evaluations = ' + myokit.float.str(rbound[3]))
                txt.append(str(e))
                try:
                    self._model.evaluate_derivatives(rstate)
                except myokit.NumericalError as en:
                    txt.append(str(en))
                raise myokit.SimulationError('\n'.join(txt))
            except Exception as e:

                # Store error state
                self._error_state = list(rstate)

                # Check for known CVODE errors
                if 'Function CVode()' in str(e):
                    raise myokit.SimulationError(str(e))

                # Check for zero step error
                if str(e)[:10] == 'ZERO_STEP ':  # pragma: no cover
                    t = float(str(e)[10:])
                    raise myokit.SimulationError(
                        'Maximum number of zero-size steps made at t='
                        + str(t))

                # Unknown exception: re-raise!
                raise
            finally:
                # Clean even after KeyboardInterrupt or other Exception
                self._sim.sim_clean()

            # Update internal state
            self._state = rstate

        # Return
        if root_list is not None:
            # Calculate apds and return (log, apds)
            st = []
            dr = []
            if root_list:
                roots = iter(root_list)
                time, direction = next(roots)
                tlast = time if direction > 0 else None
                for time, direction in roots:
                    if direction > 0:
                        tlast = time
                    else:
                        st.append(tlast)
                        dr.append(time - tlast)
            apds = myokit.DataLog()
            apds['start'] = st
            apds['duration'] = dr
            return log, apds
        else:
            # Return log
            return log

    def set_constant(self, var, value):
        """
        Changes a model constant. Only literal constants (constants not
        dependent on any other variable) can be changed.

        The constant ``var`` can be given as a :class:`Variable` or a string
        containing a variable qname. The ``value`` should be given as a float.
        """
        value = float(value)
        if isinstance(var, myokit.Variable):
            var = var.qname()
        var = self._model.get(var)
        if not var.is_literal():
            raise ValueError(
                'The given variable <' + var.qname() + '> is not a literal.')

        # Update value in internal model: This is required for error handling
        # (to show the correct values), but also takes care of constants in
        # pickled/unpickled simulations.
        self._model.set_value(var.qname(), value)

        # Update value in compiled simulation module
        self._sim.set_constant(var.qname(), value)

    def set_default_state(self, state):
        """
        Allows you to manually set the default state.
        """
        self._default_state = self._model.map_to_state(state)

    def set_max_step_size(self, dtmax=None):
        """
        Sets a maximum step size. To let the solver pick any step size it likes
        use ``dtmax = None``.
        """
        dtmax = 0 if dtmax is None else float(dtmax)
        if dtmax < 0:
            dtmax = 0

        # Store internally
        self._dtmax = dtmax

        # Set in simulation
        self._sim.set_max_step_size(dtmax)

    def set_min_step_size(self, dtmin=None):
        """
        Sets a minimum step size. To let the solver pick any step size it likes
        use ``dtmin = None``.
        """
        dtmin = 0 if dtmin is None else float(dtmin)
        if dtmin < 0:
            dtmin = 0

        # Store internally
        self._dtmin = dtmin

        # Set in simulation
        self._sim.set_min_step_size(dtmin)

    def set_fixed_form_protocol(self, times=None, values=None):
        """
        Configures this simulation to run with a predetermined protocol
        instead of the usual event-based mechanism.

        A 1D time-series should be given as input. During the simulation, the
        value of the pacing variable will be determined by linearly
        interpolating between the two nearest points in the series. If the
        simulation time is outside the bounds of the time-series, the first or
        last value in the series will be used.

        Setting a predetermined protocol clears any previously set (event-based
        or pre-determined) protocol. To clear all protocols, call this method
        with `times=None`. When a simulation is run without any protocol, the
        value of any variables bound to `pace` will be set to 0.

        Arguments:

        ``times``
            A non-decreasing array of times. If any times appear more than
            once, only the value at the highest index will be used.
        ``values``
            An array of values for the pacing variable. Must have the same size
            as ``times``.

        """
        # Check input
        if times is None:
            if values is not None:
                raise ValueError('Values array given, but no times array.')
        else:
            if values is None:
                raise ValueError('Times array given, but no values array.')
            if len(times) != len(values):
                raise ValueError('Times and values array must have same size.')

        # Clear event-based protocol, if set
        self._protocol = None

        # Set new protocol
        if times is None:
            # Clear predetermined protocol
            self._fixed_form_protocol = None
        else:
            # Copy data and set
            self._fixed_form_protocol = (list(times), list(values))

    def set_protocol(self, protocol=None):
        """
        Sets the pacing :class:`Protocol` used by this simulation.

        To run without pacing call this method with ``protocol = None``. In
        this case, the value of any variables bound to `pace` will be set to 0.
        """
        # Clear predetermined protocol, if set
        self._fixed_form_protocol = None

        # Set new protocol
        if protocol is None:
            self._protocol = None
        else:
            self._protocol = protocol.clone()

    def __setstate__(self, state):
        """
        Called after unpickling.

        See: https://docs.python.org/3/library/pickle.html#object.__setstate__
        """
        self._time = state[0]
        self._state = state[1]
        self._default_state = state[2]
        self._fixed_form_protocol = state[3]

        # The following properties need to be set on the internal simulation
        # object
        self.set_tolerance(*state[4])
        self.set_min_step_size(state[5])
        self.set_max_step_size(state[6])

    def set_state(self, state):
        """
        Sets the current state.
        """
        self._state = self._model.map_to_state(state)

    def set_time(self, time=0):
        """
        Sets the current simulation time.
        """
        self._time = float(time)

    def set_tolerance(self, abs_tol=1e-6, rel_tol=1e-4):
        """
        Sets the solver tolerances. Absolute tolerance is set using
        ``abs_tol``, relative tolerance using ``rel_tol``. For more information
        on these values, see the Sundials CVODE documentation.
        """
        abs_tol = float(abs_tol)
        if abs_tol <= 0:
            raise ValueError('Absolute tolerance must be positive float.')
        rel_tol = float(rel_tol)
        if rel_tol <= 0:
            raise ValueError('Relative tolerance must be positive float.')

        # Store tolerance in Python (for pickling)
        self._tolerance = (abs_tol, rel_tol)

        # Set tolerance in simulation
        self._sim.set_tolerance(abs_tol, rel_tol)

    def state(self):
        """
        Returns the current state.
        """
        return list(self._state)

    def time(self):
        """
        Returns the current simulation time.
        """
        return self._time



#!/usr/bin/env python3
#
# Tests the PSimulation class.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from __future__ import absolute_import, division
from __future__ import print_function, unicode_literals

import os
import unittest

import myokit
import numpy as np

from myokit.tests import DIR_DATA, CancellingReporter, WarningCollector

# Unit testing in Python 2 and 3
try:
    unittest.TestCase.assertRaisesRegex
except AttributeError:
    unittest.TestCase.assertRaisesRegex = unittest.TestCase.assertRaisesRegexp


class PSimulationTest(unittest.TestCase):
    """
    Tests the PSimulation.
    """
    def test_simple(self):

        # Load model
        m = os.path.join(DIR_DATA, 'lr-1991.mmt')
        m, p, x = myokit.load(m)

        # Create simulation
        with WarningCollector() as c:
            s = myokit.PSimulation(
                m, p, variables=['membrane.V'],
                parameters=['ina.gNa', 'ica.gCa'])
        self.assertIn('`PSimulation` is deprecated', c.text())

        # Test state & default state
        self.assertEqual(s.state(), s.default_state())

        # Test derivatives method
        dp = s.derivatives()
        self.assertEqual(dp.shape, (8, 2))
        self.assertTrue(np.all(dp == 0))

        # Run a tiny simulation
        self.assertEqual(s.time(), 0)
        s.set_step_size(0.002)
        d, dp = s.run(10, log_interval=2)
        self.assertEqual(s.time(), 10)

        # Test derivatives method
        dp = s.derivatives()
        self.assertEqual(dp.shape, (8, 2))
        self.assertFalse(np.all(dp == 0))

        # Test state & default state before & after reset
        self.assertNotEqual(s.state(), s.default_state())
        s.reset()
        self.assertEqual(s.state(), s.default_state())
        self.assertEqual(s.time(), 0)

        # Test derivatives method after reset
        dp = s.derivatives()
        self.assertEqual(dp.shape, (8, 2))
        self.assertTrue(np.all(dp == 0))

        # Test pre-pacing --> Not implemented!
        #s.pre(2)
        #self.assertEqual(s.state(), s.default_state())

        # Create without variables or parameters
        with WarningCollector() as c:
            self.assertRaisesRegex(
                ValueError, 'variables', myokit.PSimulation, m, p,
                parameters=['ina.gNa'])
            self.assertRaisesRegex(
                ValueError, 'parameters', myokit.PSimulation, m, p,
                variables=['membrane.V'])

        # Run without validated model
        m2 = m.clone()
        m2.add_component('bert')
        with WarningCollector() as c:
            s = myokit.PSimulation(
                m2, p, variables=['membrane.V'],
                parameters=['ina.gNa', 'ica.gCa'])
        s.set_step_size(0.002)
        d, dp = s.run(10, log_interval=2)

        # Variable or parameter given twice
        with WarningCollector() as c:
            self.assertRaisesRegex(
                ValueError, 'Duplicate variable', myokit.PSimulation, m, p,
                variables=['membrane.V', 'membrane.V'], parameters=['ina.gNa'])
            self.assertRaisesRegex(
                ValueError, 'Duplicate parameter', myokit.PSimulation, m, p,
                variables=['membrane.V'], parameters=['ina.gNa', 'ina.gNa'])

            # Bound variable or parameter
            self.assertRaisesRegex(
                ValueError, 'bound', myokit.PSimulation, m, p,
                variables=['engine.pace'], parameters=['ina.gNa'])
            self.assertRaisesRegex(
                ValueError, 'bound', myokit.PSimulation, m, p,
                variables=['membrane.V'], parameters=['engine.pace'])

            # Constant variable
            self.assertRaisesRegex(
                ValueError, 'constant', myokit.PSimulation, m, p,
                variables=['ica.gCa'], parameters=['ina.gNa'])

            # Non-constant parameter
            self.assertRaisesRegex(
                ValueError, 'literal constant', myokit.PSimulation, m, p,
                variables=['membrane.V'], parameters=['cell.RTF'])

            # Variables given as objects
            myokit.PSimulation(
                m, p, variables=[m.get('membrane.V')],
                parameters=[m.get('ina.gNa')])

        # Negative times
        self.assertRaisesRegex(
            ValueError, 'negative', s.run, -1)

        # Negative or zero step size
        self.assertRaisesRegex(
            ValueError, 'zero', s.set_step_size, 0)
        self.assertRaisesRegex(
            ValueError, 'zero', s.set_step_size, -1)

        # Set unset protocol
        s.set_protocol(None)
        s.set_protocol(p)

    def test_block(self):
        # Test :meth:`PSimulation.block()`.

        m, p, x = myokit.load(os.path.join(DIR_DATA, 'lr-1991.mmt'))
        with WarningCollector() as c:
            s = myokit.PSimulation(
                m, p, variables=['membrane.V'],
                parameters=['ina.gNa', 'ica.gCa'])
        s.set_step_size(0.002)
        d, dp = s.run(10, log_interval=2)

        b = s.block(d, dp)
        self.assertIsInstance(b, myokit.DataBlock2d)
        self.assertEqual(b.len0d(), len(d) - 1)
        self.assertTrue(np.all(b.time() == d.time()))

        # Log without time
        e = myokit.DataLog(d)
        del(e[e.time_key()])
        self.assertRaisesRegex(ValueError, 'must contain', s.block, e, dp)

        # Wrong size derivatives array
        self.assertRaisesRegex(ValueError, 'shape', s.block, d, dp[:, :-1])

    def test_set_constant(self):
        # Test :meth:`PSimulation.set_constant()` and
        # :meth:`PSimulation.set_parameters()`

        m, p, x = myokit.load(os.path.join(DIR_DATA, 'lr-1991.mmt'))
        with WarningCollector() as c:
            s = myokit.PSimulation(
                m, p, variables=['membrane.V'], parameters=['ina.gNa'])
        s.set_constant('ica.gCa', 1)
        s.set_constant(m.get('ica.gCa'), 1)

        # Variable is not a literal
        self.assertRaisesRegex(
            ValueError, 'literal', s.set_constant, 'membrane.V', 1)

        # Variable is in parameters list
        self.assertRaisesRegex(
            ValueError, 'parameter', s.set_constant, 'ina.gNa', 1)

        # Set parameter values
        s.set_parameters([1])
        self.assertRaisesRegex(
            ValueError, '1 values', s.set_parameters, [1, 2])
        s.set_parameters({'ina.gNa': 1})
        s.set_parameters({m.get('ina.gNa'): 1})
        self.assertRaisesRegex(
            ValueError, 'Unknown', s.set_parameters, {'bert': 2})
        self.assertRaisesRegex(
            ValueError, 'parameter', s.set_parameters, {'ica.gCa': 2})

    def test_progress_reporter(self):
        # Test running with a progress reporter.

        m, p, x = myokit.load(os.path.join(DIR_DATA, 'lr-1991.mmt'))
        with WarningCollector() as c:
            s = myokit.PSimulation(
                m, p, variables=['membrane.V'], parameters=['ina.gNa'])
        with myokit.tools.capture() as c:
            s.run(2, progress=myokit.ProgressPrinter())
        c = c.text().splitlines()
        self.assertEqual(len(c), 2)
        self.assertEqual(
            c[0], '[0.0 minutes] 50.0 % done, estimated 0 seconds remaining')
        self.assertEqual(
            c[1], '[0.0 minutes] 100.0 % done, estimated 0 seconds remaining')

        # Not a progress reporter
        self.assertRaisesRegex(
            ValueError, 'ProgressReporter', s.run, 1, progress=12)

        # Cancel from reporter
        self.assertRaises(
            myokit.SimulationCancelledError, s.run, 1,
            progress=CancellingReporter(0))


if __name__ == '__main__':
    unittest.main()



#
# Simulation that integrates the Jacobian to obtain the partial derivatives of
# the state vector with respect to the initial conditions.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from __future__ import absolute_import, division
from __future__ import print_function, unicode_literals

import os
import myokit
import numpy as np
import platform
import warnings

# Location of C template
SOURCE_FILE = 'icsim.cpp'


class ICSimulation(myokit.CppModule):
    """
    Runs a forward-Euler based simulation and calculates the partial
    derivatives of the state vector with respect to the initial conditions.

    This class is deprecated. Sensitivities with respect to initial conditions
    can now be calculated with the :class:`Simulation` class.

    The simulation is based on automatic differentiation implemented using a
    C++ data type that replaces a single scalar float with a float and a list
    of partial derivatives. Any operations on this pair update both the float
    and the set of derivatives. A normal simulation starts with a state
    ``y(tmin)`` and a right-hand side function (RHS) ``f(y) = dy/dt``. It then
    integrates ``f(y)`` from ``tmin`` to ``tmax`` resulting in an output state
    ``y(tmax)``. In this simulation the data type of ``f`` is replaced by a
    ``(f, df/dy)``, where ``df/dy`` is the matrix of partial derivatives of
    ``f`` with respect to ``y``. By integrating ``f`` from ``tmin`` to ``tmax``
    we obtain the state at ``tmax``. This can be seen as a function
    ``F(y(tmin))``, that gives the state at ``tmax`` given ``y(tmin)``. By
    integrating ``df/dy`` the derivative of ``F`` to ``y(tmin)`` is obtained.
    This result allows the sensitivity of the system to its initial conditions
    to be evaluated.

    N.B. The partial derivatives can not be calculated for the following
    functions: ``floor``, ``ceil``, ``abs``, quotients and remainders. If these
    are encountered the resulting derivatives will be yielded as ``NaN``.
    However, in many cases, these functions will only occur as part of a
    condition in an if statement, so the ``NaN``'s won't propagate to the final
    result.

    The model passed to the simulation is cloned and stored internally, so
    changes to the original model object will not affect the simulation.

    A protocol can be passed in as ``protocol`` or set later using
    :meth:`set_protocol`.

    Simulations maintain an internal state consisting of

    - the current simulation time
    - the current state
    - the derivatives of the current state with respect to the initial state

    When a simulation is created, the simulation time is set to 0 and the
    state is obtained from the given model. The initial derivatives matrix is
    an identity matrix of size ``(n, n)``, where ``n`` is the number of states
    in the model.
    After each call to :meth:`run` the time, state and derivative variables are
    updated so that each successive call to run continues where the previous
    one left off. A :meth:`reset` method is provided that will set the
    time back to 0, revert the current state to the default state and set the
    derivatives back to ``I``.

    The simulation provides two inputs a variable can bind to:

    ``time``
        This variable contains the simulation time.
    ``pace``
        This variable contains the current value of the pacing variable
        as given by the protocol passed to the Simulation.

    No labeled variables are required.
    """
    _index = 0  # Simulation id

    def __init__(self, model, protocol=None):
        super(ICSimulation, self).__init__()

        # Deprecated on 2021-02-25
        warnings.warn(
            'The class `ICSimulation` is deprecated. Sensitivities with'
            ' respect to initial conditions can now be calculated with the'
            ' single cell myokit.Simulation class.')

        # Require a valid model
        if not model.is_valid():
            model.validate()
        model = model.clone()
        self._model = model

        # Set protocol
        self.set_protocol(protocol)

        # Get state and default state from model
        self._state = self._model.state()
        self._default_state = list(self._state)

        # Create initial list of derivatives
        n = len(self._state)
        self._deriv = [0.0] * n**2
        for i in range(n):
            self._deriv[i * (n + 1)] = 1.0

        # Starting time
        self._time = 0

        # Default time step
        self._dt = 0
        self.set_step_size()

        # Unique simulation id
        ICSimulation._index += 1
        module_name = 'myokit_ICSimulation_' + str(ICSimulation._index)
        module_name += '_' + str(myokit.pid_hash())

        # Arguments
        args = {
            'module_name': module_name,
            'model': self._model,
        }
        fname = os.path.join(myokit.DIR_CFUNC, SOURCE_FILE)

        # Define libraries
        libs = []
        if platform.system() != 'Windows':  # pragma: no windows cover
            libs.append('m')

        # Create simulation
        libd = []
        incd = [myokit.DIR_CFUNC]
        self._sim = self._compile(module_name, fname, args, libs, libd, incd)

    def block(self, log, derivatives):
        """
        Takes the output of a simulation (a simulation log and a list of
        derivatives) and combines it into a single :class:`DataBlock2d` object.

        Each entry in the log is converted to a 0d entry in the log. The
        calculated derivatives are stored as the 2d field ``derivatives``.
        """
        # Get time data
        tvar = self._model.time().qname()
        try:
            time = log[tvar]
        except KeyError:
            raise ValueError(
                'The given log must contain an entry for <' + tvar + '>.')

        # Check shape of derivatives array
        n = self._model.count_states()
        shape = (len(time), n, n)
        if derivatives.shape != shape:
            raise ValueError(
                'Wrong input: Expecting a derivatives array of shape '
                + str(shape) + '.')

        # Create datablock
        block = myokit.DataBlock2d(n, n, time)
        for k, v in log.items():
            if k != tvar:
                block.set0d(k, v)
        block.set2d('derivatives', derivatives)
        return block

    def default_state(self):
        """
        Returns the default state.
        """
        return list(self._default_state)

    def derivatives(self):
        """
        Return the partial derivatives of the current state with respect to the
        initial state.
        """
        n = len(self._state)
        return np.array(self._deriv, copy=True).reshape((n, n))

    def reset(self):
        """
        Resets the simulation:

        - The time variable is set to 0
        - The state is set back to the default state

        """
        # Reset time
        self._time = 0

        # Reset state
        self._state = list(self._default_state)

        # Reset derivatives
        n = len(self._state)
        self._deriv = [0.0] * n**2
        for i in range(n):
            self._deriv[i * (n + 1)] = 1.0

    def run(self, duration, log=None, log_interval=5, progress=None,
            msg='Running ICSimulation'):
        """
        Runs a simulation and returns the logged results. Running a simulation
        has the following effects:

        - The internal state is updated to the last state in the simulation.
        - The simulation's time variable is updated to reflect the time
          elapsed during the simulation.

        The number of time units to simulate can be set with ``duration``.

        The variables to log can be indicated using the ``log`` argument. There
        are several options for its value:

        - ``None`` (default), to log all states.
        - An integer flag or a combination of flags. Options:
          ``myokit.LOG_NONE``, ``myokit.LOG_STATE``, ``myokit.LOG_INTER``,
          ``myokit.LOG_BOUND``.
        - A list of qnames or variable objects
        - A :class:`myokit.DataLog` obtained from a previous simulation.
          In this case, the newly logged data will be appended to the existing
          log.

        For more details on the ``log`` argument, see the function
        :meth:`myokit.prepare_log`.

        The method returns a :class:`myokit.DataLog` and a 3d numpy
        array. In the returned array, the first axis represents the time,
        the second axis is a state x and the third is a state y such that the
        point ``(t, x, y)`` represents ``dx/dy(0)`` at time t. For example, if
        ``p`` is the array of derivatives, to get the derivative of state 0
        with respect to the initial value of state 1, use ``p[:,0,1]``.

        A log entry is created every time *at least* ``log_interval`` time
        units have passed.

        To obtain feedback on the simulation progress, an object implementing
        the :class:`myokit.ProgressReporter` interface can be passed in.
        passed in as ``progress``. An optional description of the current
        simulation to use in the ProgressReporter can be passed in as `msg`.
        """
        # Simulation times
        if duration < 0:
            raise ValueError('Simulation time can\'t be negative.')
        tmin = self._time
        tmax = tmin + duration

        # Parse log argument
        log = myokit.prepare_log(
            log,
            self._model,
            if_empty=myokit.LOG_STATE + myokit.LOG_BOUND,
            allowed_classes=myokit.LOG_STATE + myokit.LOG_BOUND
            + myokit.LOG_INTER,
        )

        # Logging period (0 = disabled)
        log_interval = float(log_interval)
        if log_interval < 0:
            log_interval = 0

        # Create empty list for derivative lists
        derivs = []

        # Get progress indication function (if any)
        if progress is None:
            progress = myokit._Simulation_progress
        if progress:
            if not isinstance(progress, myokit.ProgressReporter):
                raise ValueError(
                    'The argument "progress" must be either a subclass of'
                    ' myokit.ProgressReporter or None.')

        # Run simulation
        if duration > 0:
            # Initialize
            n = len(self._state)
            state = [0] * n
            deriv = [0] * (n ** 2)
            self._sim.sim_init(
                tmin,
                tmax,
                self._dt,
                list(self._state),
                list(self._deriv),
                state,
                deriv,
                self._protocol,
                log,
                derivs,
                log_interval,
            )
            t = tmin
            try:
                if progress:
                    # Loop with feedback
                    with progress.job(msg):
                        r = 1 / duration
                        while t < tmax:
                            t = self._sim.sim_step()
                            if not progress.update(min((t - tmin) * r, 1)):
                                raise myokit.SimulationCancelledError()
                else:
                    # Loop without feedback
                    while t < tmax:
                        t = self._sim.sim_step()
            finally:
                # Clean even after KeyboardInterrupt or other Exception
                self._sim.sim_clean()
            # Update internal state
            self._state = list(state)
            self._deriv = list(deriv)
            self._time += duration
            # Convert derivatives to numpy arrays
            # Using
            #   derivs = [np.array(x).reshape(n,n) for x in derivs]
            # will create a list of views of arrays
            # to avoid the overhead of views, perhaps it's better to copy this
            # view into a new array explicitly?
            derivs = np.array([
                np.array(np.array(x).reshape(n, n), copy=True) for x in derivs
            ])

        # Return
        return log, derivs

    def set_protocol(self, protocol=None):
        """
        Changes the pacing protocol used by this simulation.
        """
        if protocol is None:
            self._protocol = None
        else:
            self._protocol = protocol.clone()

    def set_step_size(self, dt=0.01):
        """
        Sets the step size used in the forward Euler solving routine.
        """
        dt = float(dt)
        if dt <= 0:
            raise ValueError('Step size must be greater than zero.')
        self._dt = dt

    def state(self):
        """
        Returns the current state.
        """
        return list(self._state)

    def time(self):
        """
        Returns the current simulation time.
        """
        return self._time



#
# Runs a simulation with differential objects, to obtain the state and the
# partial derivatives of the state with respect to a list of parameters.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from __future__ import absolute_import, division
from __future__ import print_function, unicode_literals

import os
import myokit
import numpy as np
import platform
import warnings


# Location of C template
SOURCE_FILE = 'psim.cpp'


class PSimulation(myokit.CppModule):
    """
    Runs a forward-Euler based simulation and calculates the partial
    derivatives of the model variables with respect to a given set of
    parameters.

    This class is deprecated. Sensitivities with respect to parameters can now
    be calculated with the :class:`Simulation` class.

    The simulation is based on automatic differentiation implemented using a
    C++ data type that replaces a single scalar float with a float and a list
    of partial derivatives. Any operations on this pair update both the float
    and the set of derivatives.

    The resulting output is a set of logged variables plus a matrix of
    derivatives ``dy/dp`` where ``y`` is a non-constant variable and ``p`` is
    a constant parameter. The variables and parameters to track can be
    specified using :class:`myokit.Variable` objects or by their names. The
    parameters should be given as a list ``parameters`` while the variables
    ``y`` should be given in the list ``variables``.

    N.B. Partial derivatives can not be calculated for the functions ``floor``,
    ``ceil`` and ``abs`` or for quotients and remainders. If these are
    encountered the resulting derivatives will be yielded as ``NaN``.

    A protocol can be passed in as ``protocol`` or set later using
    :meth:`set_protocol`.

    The model and protocol passed to the simulation are cloned and stored
    internally. Any changes to the original model or protocol will not affect
    the simulation.

    Simulations maintain an internal state consisting of

      - the current simulation time
      - the current state
      - the partial derivatives of the current state with respect to the
        parameters

    When a simulation is created, the simulation time is set to 0 and the
    state is obtained from the given model. The derivatives matrix is
    initialised as a matrix of size ``(n, m)`` with a row for each of the ``n``
    states and a column for each of ``m`` parameters.

    After each call to :meth:`run` the time, state and derivative variables are
    updated so that each successive call to run continues where the previous
    one left off. A :meth:`reset` method is provided that will set the time
    back to 0, revert the current state to the default state and reset the
    calculated derivatives.

    The simulation provides two inputs a variable can bind to:

    ``time``
        This variable contains the simulation time.
    ``pace``
        This variable contains the current value of the pacing variable
        as given by the protocol passed to the Simulation.

    No labeled variables are required.
    """
    _index = 0  # Unique id for generated module

    def __init__(
            self, model, protocol=None, variables=None, parameters=None):
        super(PSimulation, self).__init__()

        # Deprecated on 2021-02-25
        warnings.warn(
            'The class `PSimulation` is deprecated. Sensitivities with'
            ' respect to parameters can now be calculated with the single cell'
            ' myokit.Simulation class.')

        # Check presence of variables and parameters arguments (are required
        # arguments but protocol is not...)
        if variables is None:
            raise ValueError(
                'Please specify a set of variables whose derivatives should be'
                ' tracked.')
        if parameters is None:
            raise ValueError('Please specify a set of parameters.')

        # Require a valid model
        if not model.is_valid():
            model.validate()
        model = model.clone()
        self._model = model

        # Set protocol
        self.set_protocol(protocol)

        # Check tracked variables
        if len(variables) != len(set(variables)):
            raise ValueError('Duplicate variable in `variables`.')
        self._variables = []
        for v in variables:
            if isinstance(v, myokit.Variable):
                v = v.qname()
            v = self._model.get(v, myokit.Variable)
            if not (v.is_state() or v.is_intermediary()):
                if v.is_bound():
                    raise ValueError(
                        'Variables to track cannot be bound to external'
                        ' inputs.')
                else:
                    raise ValueError('Variables to track cannot be constants.')
            self._variables.append(v)

        # Check parameters
        if len(parameters) != len(set(parameters)):
            raise ValueError('Duplicate parameter in `parameters`.')
        self._parameters = []
        for p in parameters:
            if isinstance(p, myokit.Variable):
                p = p.qname()
            p = self._model.get(p, myokit.Variable)
            if not p.is_literal():
                if p.is_bound():
                    raise ValueError(
                        'Parameters cannot be bound to external inputs.')
                else:
                    raise ValueError('Parameters must be literal constants.')
            self._parameters.append(p)
        del(parameters)

        # Create list of parameter values
        self._values = []
        for p in self._parameters:
            self._values.append(p.rhs().eval())

        # Get state and default state from model
        self._state = self._model.state()
        self._default_state = list(self._state)

        # Create list of state-parameter-derivatives
        ms = len(self._state)
        mp = len(self._parameters)
        self._state_ddp = [0.0] * (ms * mp)

        # Starting time
        self._time = 0

        # Default time step
        self._dt = 0
        self.set_step_size()

        # Unique simulation id
        PSimulation._index += 1
        module_name = 'myokit_PSimulation_' + str(PSimulation._index)
        module_name += '_' + str(myokit.pid_hash())

        # Arguments
        args = {
            'module_name': module_name,
            'model': self._model,
            'variables': self._variables,
            'parameters': self._parameters,
        }
        fname = os.path.join(myokit.DIR_CFUNC, SOURCE_FILE)

        # Define libraries
        libs = []
        if platform.system() != 'Windows':
            libs.append('m')

        # Create simulation
        libd = []
        incd = [myokit.DIR_CFUNC]
        self._sim = self._compile(module_name, fname, args, libs, libd, incd)

    def block(self, log, derivatives):
        """
        Takes the output of a simulation (a simulation log and a list of
        derivatives) and combines it into a single :class:`DataBlock2d` object.

        Each entry in the log is converted to a 0d entry in the block. The
        calculated derivatives are stored as the 2d field ``derivatives``.
        """
        # Get time data
        tvar = self._model.time().qname()
        try:
            time = log[tvar]
        except KeyError:
            raise ValueError(
                'The given log must contain an entry for <' + tvar + '>.')
        # Check shape of derivatives array
        n = len(self._variables)
        m = len(self._parameters)
        shape = (len(time), n, m)
        if derivatives.shape != shape:
            raise ValueError(
                'Wrong input: Expecting a derivatives array of shape '
                + str(shape) + '.')
        # Create datablock
        block = myokit.DataBlock2d(m, n, time)
        for k, v in log.items():
            if k != tvar:
                block.set0d(k, v)
        block.set2d('derivatives', derivatives)
        return block

    def default_state(self):
        """
        Returns the default state.
        """
        return list(self._default_state)

    def derivatives(self):
        """
        Return the partial derivatives of the current state with respect to the
        parameters. Only works once the simulation has been run!
        """
        ms = len(self._state)
        mp = len(self._parameters)
        return np.array(self._state_ddp, copy=True).reshape((ms, mp))

    def reset(self):
        """
        Resets the simulation:

        - The time variable is set to 0
        - The state is set back to the default state
        - The derivatives are set to zero

        """
        # Reset time
        self._time = 0

        # Reset state
        self._state = list(self._default_state)

        # Reset state-parameter-derivatives
        ms = len(self._state)
        mp = len(self._parameters)
        self._state_ddp = [0.0] * (ms * mp)

    def run(
            self, duration, log=None, log_interval=1, progress=None,
            msg='Running PSimulation'):
        """
        Runs a simulation and returns the logged results. Running a simulation
        has the following effects:

        - The internal state is updated to the last state in the simulation.
        - The simulation's time variable is updated to reflect the time
          elapsed during the simulation.

        The number of time units to simulate can be set with ``duration``.

        The variables to log can be indicated using the ``log`` argument. There
        are several options for its value:

        - ``None`` (default), to log all states.
        - An integer flag or a combination of flags. Options:
          ``myokit.LOG_NONE``, ``myokit.LOG_STATE``, ``myokit.LOG_INTER``,
          ``myokit.LOG_BOUND``.
        - A list of qnames or variable objects
        - A :class:`myokit.DataLog` obtained from a previous simulation.
          In this case, the newly logged data will be appended to the existing
          log.

        For more details on the ``log`` argument, see the function
        :meth:`myokit.prepare_log`.

        The method returns a :class:`myokit.DataLog` and a 3d numpy
        array. In the returned array, the first axis represents the time,
        the second axis is a tracked variable y and the third is a parameter p
        such that the point ``(t, y, p)`` represents ``dy/dp`` at time ``t``.
        For example, if ``d`` is the array of derivatives, to get the
        derivative of variables ``0`` with respect to parameter 2, use
        ``d[:,0,2]``.

        A log entry is created every time *at least* ``log_interval`` time
        units have passed. If ``log_interval <= 0`` every step taken is logged.

        To obtain feedback on the simulation progress, an object implementing
        the :class:`myokit.ProgressReporter` interface can be passed in.
        passed in as ``progress``. An optional description of the current
        simulation to use in the ProgressReporter can be passed in as `msg`.
        """
        # Simulation times
        if duration < 0:
            raise ValueError('Simulation time can\'t be negative.')
        tmin = self._time
        tmax = tmin + duration

        # Number of states, state derivatives
        ms = len(self._state)
        mv = len(self._variables)
        mp = len(self._parameters)

        # Final state and final state-parameter-derivative output lists
        state = [0.] * ms
        state_ddp = [0.] * (ms * mp)

        # Parse log argument
        log = myokit.prepare_log(
            log,
            self._model,
            if_empty=myokit.LOG_STATE + myokit.LOG_BOUND,
            allowed_classes=myokit.LOG_STATE + myokit.LOG_BOUND
            + myokit.LOG_INTER,
        )

        # Logging period (0 = log at every step)
        log_interval = float(log_interval)
        if log_interval < 0:
            log_interval = 0

        # Create empty list for variable-parameter-derivative lists
        varab_ddp = []

        # Get progress indication function (if any)
        if progress is None:
            progress = myokit._Simulation_progress
        if progress:
            if not isinstance(progress, myokit.ProgressReporter):
                raise ValueError(
                    'The argument `progress` must be either a subclass of'
                    ' myokit.ProgressReporter or None.')

        # Run simulation
        if duration > 0:
            # Initialize
            self._sim.sim_init(
                tmin,
                tmax,
                self._dt,
                list(self._values),
                list(self._state),
                list(self._state_ddp),
                state,
                state_ddp,
                self._protocol,
                log,
                varab_ddp,
                log_interval,
            )
            t = tmin
            try:
                if progress:
                    # Loop with feedback
                    with progress.job(msg):
                        r = 1.0 / duration
                        while t < tmax:
                            t = self._sim.sim_step()
                            if not progress.update(min((t - tmin) * r, 1)):
                                raise myokit.SimulationCancelledError()
                else:
                    # Loop without feedback
                    # (But with repeated returns to Python to allow Ctrl-C etc)
                    while t < tmax:
                        t = self._sim.sim_step()
            finally:
                # Clean even after KeyboardInterrupt or other Exception
                self._sim.sim_clean()

            # Update internal state
            self._state = list(state)
            self._state_ddp = list(state_ddp)
            self._time += duration

            # Convert derivatives to numpy arrays
            varab_ddp = np.array([
                np.array(np.array(x).reshape(mv, mp), copy=True)
                for x in varab_ddp])

        # Return
        return log, varab_ddp

    def set_constant(self, var, value):
        """
        Changes a model constant. Only literal constants (constants not
        dependent on any other variable) can be changed. Constants set as
        parameters cannot be changed with this method but may be set using
        :meth:`set_parameters`.

        The constant ``var`` can be given as a :class:`Variable` or a string
        containing a variable qname. The ``value`` should be given as a float.
        """
        value = float(value)
        if isinstance(var, myokit.Variable):
            var = var.qname()
        var = self._model.get(var)
        if not var.is_literal():
            raise ValueError(
                'The given variable <' + var.qname() + '> is not a literal (it'
                ' depends on other variables)')
        if var in self._parameters:
            raise ValueError(
                'The given variable <' + var.qname() + '> is set as a'
                ' parameter. Use set_parameters() instead.')

        # Update value in compiled simulation module
        self._sim.set_constant(var.qname(), value)

        # Update value in internal model
        self._model.set_value(var.qname(), value)

    def set_parameters(self, values):
        """
        Changes the values of the parameters under investigation.

        The argument ``values`` must either be an ordered sequence containing
        the values for every parameter, or a mapping from one or more parameter
        names to their new values.

        **N.B. Calling this method will reset the simulation.**
        """
        if isinstance(values, dict):
            # Create list to update so that property change only happens after
            # error checks.
            new_values = list(self._values)

            # Check all key-value pairs
            for k, v in values.items():
                if isinstance(k, myokit.Variable):
                    k = k.qname()
                try:
                    k = self._model.get(k, myokit.Variable)
                except KeyError:
                    raise ValueError('Unknown parameter: <' + str(k) + '>.')
                try:
                    i = self._parameters.index(k)
                except ValueError:
                    raise ValueError(
                        'Variable <' + str(k) + '> was not set as a'
                        ' parameter.')
                new_values[i] = float(v)
            self._values = new_values

        else:

            # Check size of list & set
            if len(values) != len(self._values):
                raise ValueError(
                    'Argument `values` should be either a dict or a list of '
                    + str(len(self._values)) + ' values.')
            self._values = [float(x) for x in values]

        # Reset the simulation: the stored partial derivatives are no longer
        # accurate.
        self.reset()

    def set_protocol(self, protocol=None):
        """
        Changes the pacing protocol used by this simulation.
        """
        if protocol is None:
            self._protocol = None
        else:
            self._protocol = protocol.clone()

    def set_step_size(self, dt=0.01):
        """
        Sets the step size used in the forward Euler solving routine.
        """
        dt = float(dt)
        if dt <= 0:
            raise ValueError('Step size must be greater than zero.')
        self._dt = dt

    def state(self):
        """
        Returns the current state.
        """
        return list(self._state)

    def time(self):
        """
        Returns the current simulation time.
        """
        return self._time



#!/usr/bin/env python3
#
# Tests the exporters from the format module.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from __future__ import absolute_import, division
from __future__ import print_function, unicode_literals

import os
import sys
import unittest

import myokit

from myokit.tests import TemporaryDirectory


# Empty config
config1 = """
"""

# Config with lots of settings
config2 = """
[myokit]
[time]
date_format = TEST_DATE_FORMAT
time_format = TEST_TIME_FORMAT
[gui]
[sundials]
lib = one;two
inc = three;four
[opencl]
lib = five;six
inc = three;eight
"""

# Config with empty paths and spaces
config_empties_1 = """
[sundials]
lib = five
inc = three;four;
[opencl]
lib = one;;   two point five;three;
inc =
"""

# Config with empty paths and " ;", which in Python 2 is ignored
config_empties_2 = """
[sundials]
lib = five ; six
inc = three;four;
[opencl]
lib = one ;;   two point five ;three;
inc =
"""

# Qt options
config_pyside = """
[gui]
backend=pyside
"""

config_pyside2 = """
[gui]
backend=pyside2
"""

config_pyqt4 = """
[gui]
backend=pyqt4
"""

config_pyqt5 = """
[gui]
backend=pyqt5
"""

# Config with extra bits
config3 = """
[extra]
setting = 12
sotting=2
str=String value
"""


class TestConfig(unittest.TestCase):

    def test_create(self):
        # Test if the `_create` method works.

        # Import hidden _config module
        path = sys.path
        try:
            sys.path.append(myokit.DIR_MYOKIT)
            import _config as config
        finally:
            sys.path = path

        # Test _create
        with TemporaryDirectory() as d:
            filename = d.path('test.ini')
            config._create(filename)
            self.assertTrue(os.path.isfile(filename))
        self.assertFalse(os.path.isfile(filename))

    def test_load_read(self):
        # Test if the `_load` method works, when a config file exists.

        # Import hidden _config module
        path = sys.path
        try:
            sys.path.append(myokit.DIR_MYOKIT)
            import _config as config
        finally:
            sys.path = path

        # Back-up current settings
        date_format = myokit.DATE_FORMAT
        time_format = myokit.TIME_FORMAT
        force_pyside = myokit.FORCE_PYSIDE
        force_pyside2 = myokit.FORCE_PYSIDE2
        force_pyqt4 = myokit.FORCE_PYQT4
        force_pyqt5 = myokit.FORCE_PYQT5
        sundials_lib = myokit.SUNDIALS_LIB
        sundials_inc = myokit.SUNDIALS_INC
        opencl_lib = myokit.OPENCL_LIB
        opencl_inc = myokit.OPENCL_INC

        # Change myokit config dir temporarily
        path = myokit.DIR_USER
        try:
            with TemporaryDirectory() as d:
                myokit.DIR_USER = d.path()

                # Simple test
                with open(d.path('myokit.ini'), 'w') as f:
                    f.write(config1)
                config._load()

                # Full values, PySide gui
                myokit.SUNDIALS_LIB = []
                myokit.SUNDIALS_INC = []
                myokit.OPENCL_LIB = []
                myokit.OPENCL_INC = []
                myokit.FORCE_PYSIDE = myokit.FORCE_PYSIDE2 = False
                myokit.FORCE_PYQT4 = myokit.FORCE_PYQT5 = False
                with open(d.path('myokit.ini'), 'w') as f:
                    f.write(config2)
                config._load()
                self.assertEqual(myokit.DATE_FORMAT, 'TEST_DATE_FORMAT')
                self.assertEqual(myokit.TIME_FORMAT, 'TEST_TIME_FORMAT')
                self.assertFalse(myokit.FORCE_PYSIDE)
                self.assertFalse(myokit.FORCE_PYSIDE2)
                self.assertFalse(myokit.FORCE_PYQT4)
                self.assertFalse(myokit.FORCE_PYQT5)
                self.assertEqual(myokit.SUNDIALS_LIB, ['one', 'two'])
                self.assertEqual(myokit.SUNDIALS_INC, ['three', 'four'])
                self.assertEqual(myokit.OPENCL_LIB, ['five', 'six'])
                self.assertEqual(myokit.OPENCL_INC, ['three', 'eight'])

                # Lists of paths should be filtered for empty values and
                # trimmed
                myokit.SUNDIALS_LIB = []
                myokit.SUNDIALS_INC = []
                myokit.OPENCL_LIB = []
                myokit.OPENCL_INC = []
                with open(d.path('myokit.ini'), 'w') as f:
                    f.write(config_empties_1)
                config._load()
                self.assertEqual(myokit.SUNDIALS_LIB, ['five'])
                self.assertEqual(myokit.SUNDIALS_INC, ['three', 'four'])
                self.assertEqual(
                    myokit.OPENCL_LIB,
                    ['one', 'two point five', 'three'])
                self.assertEqual(myokit.OPENCL_INC, [])

                # Even if the list contains " ;", which Python 2's config
                # parser treats as a comment
                myokit.SUNDIALS_LIB = []
                myokit.SUNDIALS_INC = []
                myokit.OPENCL_LIB = []
                myokit.OPENCL_INC = []
                with open(d.path('myokit.ini'), 'w') as f:
                    f.write(config_empties_2)
                if sys.hexversion < 0x03020000:
                    self.assertRaises(ImportError, config._load)
                else:
                    config._load()
                    self.assertEqual(myokit.SUNDIALS_LIB, ['five', 'six'])
                    self.assertEqual(myokit.SUNDIALS_INC, ['three', 'four'])
                    self.assertEqual(
                        myokit.OPENCL_LIB,
                        ['one', 'two point five', 'three'])
                    self.assertEqual(myokit.OPENCL_INC, [])

                # Qt gui options
                with open(d.path('myokit.ini'), 'w') as f:
                    f.write(config_pyqt4)
                config._load()
                self.assertFalse(myokit.FORCE_PYSIDE)
                self.assertFalse(myokit.FORCE_PYSIDE2)
                self.assertTrue(myokit.FORCE_PYQT4)
                self.assertFalse(myokit.FORCE_PYQT5)

                with open(d.path('myokit.ini'), 'w') as f:
                    f.write(config_pyqt5)
                config._load()
                self.assertFalse(myokit.FORCE_PYSIDE)
                self.assertFalse(myokit.FORCE_PYSIDE2)
                self.assertFalse(myokit.FORCE_PYQT4)
                self.assertTrue(myokit.FORCE_PYQT5)

                with open(d.path('myokit.ini'), 'w') as f:
                    f.write(config_pyside)
                config._load()
                self.assertTrue(myokit.FORCE_PYSIDE)
                self.assertFalse(myokit.FORCE_PYSIDE2)
                self.assertFalse(myokit.FORCE_PYQT4)
                self.assertFalse(myokit.FORCE_PYQT5)

                with open(d.path('myokit.ini'), 'w') as f:
                    f.write(config_pyside2)
                config._load()
                self.assertFalse(myokit.FORCE_PYSIDE)
                self.assertTrue(myokit.FORCE_PYSIDE2)
                self.assertFalse(myokit.FORCE_PYQT4)
                self.assertFalse(myokit.FORCE_PYQT5)

                # Odd ini file
                with open(d.path('myokit.ini'), 'w') as f:
                    f.write(config3)
                config._load()

                # No ini file (calls _create())
                os.remove(d.path('myokit.ini'))
                config._load()

        finally:
            # Reset path
            myokit.DIR_USER = path

            # Reset data and time
            myokit.DATE_FORMAT = date_format
            myokit.TIME_FORMAT = time_format
            myokit.FORCE_PYSIDE = force_pyside
            myokit.FORCE_PYSIDE2 = force_pyside2
            myokit.FORCE_PYQT4 = force_pyqt4
            myokit.FORCE_PYQT5 = force_pyqt5
            myokit.SUNDIALS_LIB = sundials_lib
            myokit.SUNDIALS_INC = sundials_inc
            myokit.OPENCL_LIB = opencl_lib
            myokit.OPENCL_INC = opencl_inc

            # Reload local settings
            config._load()

            # Sanity check
            self.assertNotEqual(myokit.DATE_FORMAT, 'TEST_DATE_FORMAT')
            self.assertNotEqual(myokit.TIME_FORMAT, 'TEST_TIME_FORMAT')
            self.assertEqual(myokit.FORCE_PYSIDE, force_pyside)
            self.assertEqual(myokit.FORCE_PYSIDE2, force_pyside2)
            self.assertEqual(myokit.FORCE_PYQT4, force_pyqt4)
            self.assertEqual(myokit.FORCE_PYQT5, force_pyqt5)
            self.assertNotEqual(myokit.SUNDIALS_LIB, ['one', 'two'])
            self.assertNotEqual(myokit.SUNDIALS_INC, ['three', 'four'])
            self.assertNotEqual(myokit.OPENCL_LIB, ['five', 'six'])
            self.assertNotEqual(myokit.OPENCL_INC, ['three', 'eight'])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Tests the Compiler detection class.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from __future__ import absolute_import, division
from __future__ import print_function, unicode_literals

import unittest

import myokit


# Strings in Python2 and Python3
try:
    basestring
except NameError:   # pragma: no cover
    basestring = str


class CompilerDetectionTest(unittest.TestCase):
    """
    Tests the compiler detection.
    """
    def test_compiler(self):
        # Test the compiler detection.
        self.assertIsInstance(myokit.Compiler.info(), basestring)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Tests the CVODE simulation class.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from __future__ import absolute_import, division
from __future__ import print_function, unicode_literals

import os
import pickle
import platform
import re
import sys
import unittest

import numpy as np

import myokit

from myokit.tests import DIR_DATA, CancellingReporter, WarningCollector

# Unit testing in Python 2 and 3
try:
    unittest.TestCase.assertRaisesRegex
except AttributeError:
    unittest.TestCase.assertRaisesRegex = unittest.TestCase.assertRaisesRegexp


@unittest.skipIf(platform.system() != 'Linux', 'Legacy CVODE tests')
class LegacySimulationTest(unittest.TestCase):
    """
    Tests the Legacy CVODE simulation class.
    """

    @classmethod
    def setUpClass(cls):
        # Test simulation creation.

        m, p, x = myokit.load(os.path.join(DIR_DATA, 'lr-1991.mmt'))
        cls.model = m
        cls.protocol = p
        cls.sim = myokit.LegacySimulation(cls.model, cls.protocol)

    def test_pre(self):
        # Test pre-pacing.

        self.sim.reset()
        self.sim.pre(200)

    def test_simple(self):
        # Test simple run.

        self.sim.reset()
        self.assertEqual(self.sim.time(), 0)
        self.sim.pre(5)
        self.assertEqual(self.sim.time(), 0)
        d = self.sim.run(5)
        self.assertEqual(self.sim.time(), 5)
        self.sim.set_time(0)
        self.assertEqual(self.sim.time(), 0)
        self.assertEqual(type(d), myokit.DataLog)
        self.assertIn('engine.time', d)
        n = len(d['engine.time'])
        for k, v in d.items():
            self.assertEqual(n, len(v))

        # Can't do negative times
        self.assertRaisesRegex(ValueError, 'negative', self.sim.run, -1)

        # Negative log interval is set to zero
        self.sim.reset()
        d1 = self.sim.run(5)
        self.sim.reset()
        d2 = self.sim.run(5, log_interval=-5)
        self.assertEqual(d1.time(), d2.time())

    def test_no_protocol(self):
        # Test running without a protocol.

        self.sim.reset()
        self.sim.pre(50)
        self.sim.set_protocol(None)
        d = self.sim.run(50).npview()

        # Check if pace was set to zero (see prop 651 / technical docs).
        self.assertTrue(np.all(d['engine.pace'] == 0.0))

    def test_fixed_form_protocol(self):
        # Test running with a fixed form protocol.

        n = 10
        time = list(range(n))
        pace = [0] * n
        pace[2:4] = [0.5, 0.5]

        self.sim.set_fixed_form_protocol(time, pace)
        self.sim.reset()
        d = self.sim.run(n, log_interval=1)
        self.assertEqual(list(d.time()), time)
        self.assertEqual(list(d['engine.pace']), pace)

        # Unset
        self.sim.set_fixed_form_protocol(None)
        self.sim.reset()
        d = self.sim.run(n, log_interval=1)
        self.assertEqual(list(d['engine.pace']), [0] * n)

        # Reset
        self.sim.set_fixed_form_protocol(time, pace)
        self.sim.reset()
        d = self.sim.run(n, log_interval=1)
        self.assertEqual(list(d.time()), time)
        self.assertEqual(list(d['engine.pace']), pace)

        # Unset, replace with original protocol
        self.sim.set_protocol(self.protocol)
        self.sim.reset()
        d = self.sim.run(n, log_interval=1)
        self.assertNotEqual(list(d['engine.pace']), pace)
        self.assertNotEqual(list(d['engine.pace']), [0] * n)

        # Invalid protocols
        self.assertRaisesRegex(
            ValueError, 'no times', self.sim.set_fixed_form_protocol,
            values=pace)
        self.assertRaisesRegex(
            ValueError, 'no values', self.sim.set_fixed_form_protocol,
            times=time)
        self.assertRaisesRegex(
            ValueError, 'same size', self.sim.set_fixed_form_protocol,
            time, pace[:-1])

    def test_in_parts(self):
        # Test running the simulation in parts.

        self.sim.reset()
        # New logs should start with first state, finish with final
        d = self.sim.run(150)
        self.assertEqual(d['engine.time'][0], 0.0)
        self.assertEqual(d['engine.time'][-1], 150.0)
        # Next part should continue at 150, leave where last left off
        e = self.sim.run(50)
        self.assertEqual(d['engine.time'][-1], e['engine.time'][0])
        self.assertEqual(d['membrane.V'][-1], e['membrane.V'][0])
        # Re-used logs shouldn't re-log their first state
        n = len(e['engine.time'])
        e = self.sim.run(50, log=e)
        self.assertNotEqual(e['engine.time'][n - 1], e['engine.time'][n])
        self.assertGreater(e['engine.time'][n], e['engine.time'][n - 1])

    def test_pacing_values_at_event_transitions(self):
        # Tests the value of the pacing signal at event transitions

        # Create a simple model
        m = myokit.Model()
        c = m.add_component('c')
        t = c.add_variable('t')
        t.set_rhs(0)
        t.set_binding('time')
        v = c.add_variable('v')
        v.set_rhs('0')
        v.set_binding('pace')
        x = c.add_variable('x')
        x.set_rhs(0.1)
        x.promote(0)

        # Create step protocol
        p = myokit.Protocol()
        p.schedule(0, 0, 2)
        p.schedule(1, 2, 2)
        p.schedule(2, 4, 4)
        p.schedule(3, 8, 2)

        # Simulate with dynamic logging
        s = myokit.LegacySimulation(m, p)
        d = s.run(p.characteristic_time())
        time = list(d.time())
        value = list(d['c.v'])

        if False:
            for i, t in enumerate(d.time()):
                t = str(np.round(t, 5))
                print(t + ' ' * (10 - len(t)) + str(d['c.v'][i]))

        # Values should be
        #   t   0   1   2   3   4   5   6   7   8   9   10
        #   p   0   0   1   1   2   2   2   2   3   3   0
        self.assertEqual(value[time.index(0.0)], 0)
        self.assertEqual(value[time.index(0.0) + 1], 0)
        self.assertEqual(value[time.index(2.0) - 1], 0)
        self.assertEqual(value[time.index(2.0)], 1)
        self.assertEqual(value[time.index(2.0) + 1], 1)
        self.assertEqual(value[time.index(4.0) - 1], 1)
        self.assertEqual(value[time.index(4.0)], 2)
        self.assertEqual(value[time.index(4.0) + 1], 2)
        self.assertEqual(value[time.index(8.0) - 1], 2)
        self.assertEqual(value[time.index(8.0)], 3)
        self.assertEqual(value[time.index(8.0) + 1], 3)
        self.assertEqual(value[time.index(10.0) - 1], 3)
        self.assertEqual(value[time.index(10.0)], 0)

        # Simulate with fixed logging
        s.reset()
        d = s.run(p.characteristic_time() + 1, log_times=d.time())
        time2 = list(d.time())
        value2 = list(d['c.v'])
        self.assertEqual(time, time2)
        self.assertEqual(value, value2)

    def test_progress_reporter(self):
        # Test running with a progress reporter.

        # Test if it works
        sim = myokit.LegacySimulation(self.model, self.protocol)
        with myokit.tools.capture() as c:
            sim.run(110, progress=myokit.ProgressPrinter())
        c = c.text().splitlines()
        self.assertEqual(len(c), 2)
        p = re.compile(re.escape('[0.0 minutes] 1.9 % done, estimated ') +
                       '[0-9]+' + re.escape(' seconds remaining'))
        self.assertIsNotNone(p.match(c[0]))
        p = re.compile(re.escape('[0.0 minutes] 100.0 % done, estimated ') +
                       '[0-9]+' + re.escape(' seconds remaining'))
        self.assertIsNotNone(p.match(c[1]))

        # Not a progress reporter
        self.assertRaisesRegex(
            ValueError, 'ProgressReporter', self.sim.run, 5, progress=12)

        # Cancel from reporter
        self.assertRaises(
            myokit.SimulationCancelledError, self.sim.run, 1,
            progress=CancellingReporter(0))

    def test_apd_tracking(self):
        # Test the APD calculation method.

        # More testing is done in test_datalog.py!

        # Apd var is not a state
        v = self.model.get('ina.INa')
        self.assertRaisesRegex(
            ValueError, 'must be a state', myokit.LegacySimulation, self.model,
            self.protocol, apd_var=v)

        # Set a valid apd variable
        v = self.model.get('ik.x')
        sim = myokit.LegacySimulation(
            self.model, self.protocol, apd_var=v)
        sim.run(1, apd_threshold=12)

        # No apd var given, but threshold provided
        self.assertRaisesRegex(
            ValueError, 'without apd_var', self.sim.run, 1, apd_threshold=12)

    def test_last_state(self):
        # Returns the last state before an error, or None.

        m = self.model.clone()
        istim = m.get('membrane.i_stim')
        istim.set_rhs('engine.pace / stim_amplitude')
        s = myokit.LegacySimulation(m, self.protocol)
        self.assertIsNone(s.last_state())
        s.run(1)
        self.assertIsNone(s.last_state())
        s.set_constant('membrane.i_stim.stim_amplitude', 0)
        s.reset()
        self.assertRaisesRegex(myokit.SimulationError, "at t = 0", s.run, 5)
        self.assertEqual(len(s.last_state()), len(s.state()))
        self.assertEqual(s.last_state(), s.state())

    def test_last_evaluations_and_steps(self):
        # Test :meth:`LegacySimulation.last_number_of_evaluations()` and
        # :meth:`LegacySimulation.last_number_of_steps()`

        s = myokit.LegacySimulation(self.model, self.protocol)
        self.assertEqual(s.last_number_of_evaluations(), 0)
        self.assertEqual(s.last_number_of_steps(), 0)
        s.run(1)
        self.assertTrue(s.last_number_of_evaluations() > 0)
        self.assertTrue(s.last_number_of_steps() > 0)
        self.assertNotEqual(
            s.last_number_of_evaluations(), s.last_number_of_steps())

    def test_eval_derivatives(self):
        # Test :meth:`LegacySimulation.eval_derivatives()`.

        self.sim.reset()
        s1 = self.sim.state()
        d1 = self.sim.eval_derivatives()
        self.sim.run(1)
        d2 = self.sim.eval_derivatives()
        self.assertNotEqual(d1, d2)
        self.assertEqual(d1, self.sim.eval_derivatives(s1))
        self.sim.set_state(s1)
        self.assertEqual(d1, self.sim.eval_derivatives())

    def test_set_tolerance(self):
        # Test :meth:`LegacySimulation.set_tolerance()`.

        self.assertRaisesRegex(
            ValueError, 'Absolute', self.sim.set_tolerance, abs_tol=0)
        self.assertRaisesRegex(
            ValueError, 'Relative', self.sim.set_tolerance, rel_tol=0)
        self.sim.set_tolerance(1e-6, 1e-4)

    def test_set_step_size(self):
        # Test :meth:`LegacySimulation.set_min_step_size()` and
        # :meth:`LegacySimulation.set_max_step_size()`.

        # Minimum: set, unset, allow negative value to unset
        self.sim.set_min_step_size(0.1)
        self.sim.set_min_step_size(None)
        self.sim.set_min_step_size(-1)

        # Same for max
        self.sim.set_max_step_size(0.1)
        self.sim.set_max_step_size(None)
        self.sim.set_max_step_size(-1)

    def test_set_state(self):
        # Test :meth:`LegacySimulation.set_state()` and
        # :meth:`LegacySimulation.set_default_state()`.

        # Get state and default state, both different from current
        state = self.sim.state()
        state[0] += 1
        default_state = self.sim.default_state()
        default_state[1] += 1
        if state == default_state:
            default_state[0] += 2

        self.assertNotEqual(self.sim.state(), state)
        self.assertNotEqual(self.sim.default_state(), default_state)

        self.sim.set_state(state)
        self.sim.set_default_state(default_state)

        self.assertEqual(self.sim.state(), state)
        self.assertEqual(self.sim.default_state(), default_state)

    def test_set_constant(self):
        # Test :meth:`LegacySimulation.set_constant()`.

        # Literal
        v = self.model.get('cell.Na_i')
        self.sim.set_constant(v, 11)
        self.assertRaises(KeyError, self.sim.set_constant, 'cell.Bert', 11)

        # Calculated constant
        self.assertRaisesRegex(
            ValueError, 'not a literal', self.sim.set_constant, 'ina.ENa', 11)

    def test_short_runs(self):
        # Test for simulations run a very short time

        # Run for 1 unit (OK)
        self.sim.reset()
        self.sim.run(1)

        # Test running for 0 units doesn't affect state
        x0 = self.sim.state()
        self.sim.run(0)
        self.assertEqual(x0, self.sim.state())
        self.sim.run(0)
        self.assertEqual(x0, self.sim.state())

        # Test running between indistinguishable times doesn't affect state
        t = self.sim.time()
        d = 0.5 * sys.float_info.epsilon
        self.assertEqual(t, t + d)
        self.sim.run(d)

        # Test running between only just distinguishable times is fine
        self.sim.reset()
        self.sim.run(1)
        t = self.sim.time()
        d = 3 * sys.float_info.epsilon
        self.assertNotEqual(t, t + d)
        self.sim.run(d)

        # Test running between barely distinguishable times raises CVODE error.
        t = self.sim.time()
        d = 2 * sys.float_info.epsilon
        self.assertNotEqual(t, t + d)
        self.assertRaisesRegex(
            myokit.SimulationError, 'CV_TOO_CLOSE', self.sim.run, d)

        # Empty log times
        self.sim.reset()
        self.sim.run(1, log_times=[])

        # Non-monotonic times
        self.sim.reset()
        with self.assertRaisesRegex(ValueError, 'Values in log_times'):
            self.sim.run(1, log_times=[1, 2, 1])

        # Simultaneous use of log_times and log_interval
        self.sim.reset()
        with self.assertRaisesRegex(ValueError, 'The arguments log_times'):
            self.sim.run(1, log_times=[1, 2], log_interval=2)

    def test_simulation_error_1(self):
        # Test for simulation error detection: massive stimulus.

        # Silly protocol
        p = myokit.Protocol()
        p.schedule(level=1000, start=1, duration=1)
        self.sim.reset()
        self.sim.set_protocol(p)
        self.assertRaisesRegex(
            myokit.SimulationError, 'numerical error', self.sim.run, 10)
        self.sim.set_protocol(self.protocol)

    @unittest.skipIf(platform.system() != 'Linux', 'CVODE error tests')
    def test_simulation_error_2(self):
        # Test for simulation error detection: failure occurred too often.

        # Cvode error (test failure occurred too many times)
        m = self.model.clone()
        v = m.get('membrane.V')
        v.set_rhs(myokit.Multiply(v.rhs(), myokit.Number(1e18)))
        s = myokit.LegacySimulation(m, self.protocol)
        with WarningCollector():
            self.assertRaisesRegex(
                myokit.SimulationError, 'numerical error', s.run, 5000)

    def test_cvode_simulation_with_zero_states(self):
        # Tests running cvode simulations on models with no ODEs

        # Create a model without states
        m1 = myokit.Model()
        c = m1.add_component('c')
        t = c.add_variable('t')
        t.set_rhs(0)
        t.set_binding('time')
        v = c.add_variable('v')
        v.set_rhs('0')
        v.set_binding('pace')
        w = c.add_variable('w')
        w.set_rhs('2 * v')

        # Create a model with a state
        m2 = m1.clone()
        z = m2.get('c').add_variable('z')
        z.set_rhs(0.1)
        z.promote(0)

        # Test without protocol and dynamic logging
        s1 = myokit.LegacySimulation(m1)
        d1 = s1.run(5)
        self.assertEqual(len(d1.time()), 2)
        self.assertEqual(list(d1.time()), [0, 5])
        self.assertEqual(list(d1['c.w']), [0, 0])
        s2 = myokit.LegacySimulation(m2)
        d2 = s2.run(6, log_times=d1.time())
        self.assertEqual(d1.time(), d2.time())
        self.assertEqual(d1['c.w'], d2['c.w'])

        # Test with a protocol and dynamic logging
        p = myokit.Protocol()
        p.schedule(0, 0, 2)
        p.schedule(1, 2, 2)
        p.schedule(2, 4, 4)
        p.schedule(3, 8, 2)
        s1.reset()
        s1.set_protocol(p)
        d1 = s1.run(p.characteristic_time())
        self.assertEqual(len(d1.time()), 5)
        self.assertEqual(list(d1.time()), [0, 2, 4, 8, 10])
        self.assertEqual(list(d1['c.w']), [0, 2, 4, 6, 0])
        s2.reset()
        s2.set_protocol(p)
        d2 = s2.run(p.characteristic_time() + 1, log_times=d1.time())
        self.assertEqual(d1.time(), d2.time())
        self.assertEqual(d1['c.w'], d2['c.w'])

        # Test with fixed logging times
        s1.reset()
        d1 = s1.run(p.characteristic_time() + 1, log_times=d1['c.t'])
        self.assertEqual(list(d1.time()), [0, 2, 4, 8, 10])
        self.assertEqual(list(d1['c.w']), [0, 2, 4, 6, 0])
        s2.reset()
        d2 = s2.run(p.characteristic_time() + 1, log_times=d1.time())
        self.assertEqual(d1.time(), d2.time())
        self.assertEqual(d1['c.w'], d2['c.w'])

        # Test appending to log
        s1.reset()
        d1 = s1.run(5)
        d1 = s1.run(5, log=d1)
        self.assertEqual(list(d1.time()), [0, 2, 4, 5, 8, 10])
        self.assertEqual(list(d1['c.w']), [0, 2, 4, 4, 6, 0])

        # Test with a log interval
        s1.reset()
        d1 = s1.run(11, log_interval=1)
        self.assertEqual(list(d1.time()), [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        self.assertEqual(list(d1['c.w']), [0, 0, 2, 2, 4, 4, 4, 4, 6, 6, 0])

    def test_pickling(self):
        # Test pickling a simulation

        # Test with myokit.Protocol
        m, p, _ = myokit.load('example')
        s1 = myokit.LegacySimulation(m, p)
        s1.pre(123)
        s_bytes = pickle.dumps(s1)
        s2 = pickle.loads(s_bytes)
        self.assertEqual(s1.time(), s2.time())
        self.assertEqual(s1.state(), s2.state())
        self.assertEqual(s1.default_state(), s2.default_state())
        s1.run(123, log=myokit.LOG_NONE)
        s2.run(123, log=myokit.LOG_NONE)
        self.assertEqual(s1.time(), s2.time())
        self.assertEqual(s1.state(), s2.state())

        # Test simulation properties
        s1.set_tolerance(1e-8, 1e-8)
        s1.set_min_step_size(1e-2)
        s1.set_max_step_size(0.1)
        s2 = pickle.loads(pickle.dumps(s1))
        s1.run(23, log=myokit.LOG_NONE)
        s2.run(23, log=myokit.LOG_NONE)
        self.assertEqual(s1.time(), s2.time())
        self.assertEqual(s1.state(), s2.state())

        # Test changed constants
        s1.set_constant('membrane.C', 1.1)
        s2 = pickle.loads(pickle.dumps(s1))
        s1.run(17, log=myokit.LOG_NONE)
        s2.run(17, log=myokit.LOG_NONE)
        self.assertEqual(s1.time(), s2.time())
        self.assertEqual(s1.state(), s2.state())

    def test_sim_stats(self):
        # Test extraction of simulation statistics
        m, p, _ = myokit.load('example')
        rt = m['engine'].add_variable('realtime')
        rt.set_rhs(0)
        rt.set_binding('realtime')
        ev = m['engine'].add_variable('evaluations')
        ev.set_rhs(0)
        ev.set_binding('evaluations')
        s = myokit.LegacySimulation(m, p)
        d = s.run(100, log=myokit.LOG_BOUND).npview()

        self.assertIn('engine.realtime', d)
        self.assertIn('engine.evaluations', d)
        rt, ev = d['engine.realtime'], d['engine.evaluations']
        self.assertEqual(len(d.time()), len(rt))
        self.assertEqual(len(d.time()), len(ev))
        self.assertTrue(np.all(rt >= 0))
        self.assertTrue(np.all(ev >= 0))
        self.assertTrue(np.all(rt[1:] >= rt[:-1]))
        self.assertTrue(np.all(ev[1:] >= ev[:-1]))

    def test_apd(self):
        # Test the apd rootfinding routine

        s = myokit.LegacySimulation(
            self.model, self.protocol, apd_var='membrane.V')
        s.set_tolerance(1e-8, 1e-8)
        d, apds = s.run(1800, log=myokit.LOG_NONE, apd_threshold=-70)

        # Check with threshold equal to V
        self.assertEqual(len(apds['start']), 2)
        self.assertEqual(len(apds['duration']), 2)
        self.assertAlmostEqual(apds['start'][0], 1.19, places=1)
        self.assertAlmostEqual(apds['start'][1], 1001.19, places=1)
        self.assertAlmostEqual(apds['duration'][0], 383.88262, places=0)
        self.assertAlmostEqual(apds['duration'][1], 378.31448, places=0)

    def test_derivatives(self):
        # Tests logging of derivatives by comparing with a finite difference
        # approximation

        # Run past the upstroke, where finite diff approx is worst
        self.sim.reset()
        self.sim.run(52)

        # Now run logged part
        d = self.sim.run(600).npview()
        if False:
            import matplotlib.pyplot as plt
            plt.figure()
            ax = plt.subplot(3, 1, 1)
            ax.plot(d.time(), d['membrane.V'])
            ax = plt.subplot(3, 1, 2)
            ax.plot(d.time(), d['dot(membrane.V)'])

        # Get central difference approximation
        t = d.time()
        v = d['membrane.V']
        dv = (v[2:] - v[:-2]) / (t[2:] - t[:-2])
        e = d['dot(membrane.V)'][1:-1] - dv
        if False:
            t = t[1:-1]
            ax.plot(t, dv, '--')
            ax = plt.subplot(3, 1, 3)
            ax.plot(t, e)
            print(np.max(np.abs(e)))

        # Compare
        self.assertLess(np.max(np.abs(e)), 0.1)


if __name__ == '__main__':
    unittest.main()




#!/usr/bin/env python3
#
# Tests the ICSimulation class.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from __future__ import absolute_import, division
from __future__ import print_function, unicode_literals

import os
import unittest
import numpy as np

import myokit

from myokit.tests import DIR_DATA, CancellingReporter, WarningCollector

# Unit testing in Python 2 and 3
try:
    unittest.TestCase.assertRaisesRegex
except AttributeError:
    unittest.TestCase.assertRaisesRegex = unittest.TestCase.assertRaisesRegexp


class ICSimulationTest(unittest.TestCase):
    """
    Tests the :class:`ICSimulation`.
    """
    def test_basic(self):
        # Test basic usage.
        # Load model
        m, p, _ = myokit.load(os.path.join(DIR_DATA, 'lr-1991.mmt'))
        n = m.count_states()

        # Run a simulation
        with WarningCollector() as c:
            s = myokit.ICSimulation(m, p)
        self.assertIn('`ICSimulation` is deprecated', c.text())

        self.assertEqual(s.time(), 0)
        self.assertEqual(s.state(), m.state())
        self.assertEqual(s.default_state(), m.state())
        self.assertTrue(np.all(s.derivatives() == np.eye(n)))
        d, e = s.run(20, log_interval=5)
        self.assertEqual(s.time(), 20)
        self.assertNotEqual(s.state(), m.state())
        self.assertEqual(s.default_state(), m.state())
        self.assertFalse(np.all(s.derivatives() == np.eye(n)))

        # Create a datablock from the simulation log
        b = s.block(d, e)

        # Calculate eigenvalues
        b.eigenvalues('derivatives')

        # Log with missing time value
        d2 = d.clone()
        del(d2['engine.time'])
        self.assertRaisesRegex(ValueError, 'time', s.block, d2, e)

        # Wrong size derivatives array
        self.assertRaisesRegex(ValueError, 'shape', s.block, d, e[:-1])

        # Time can't be negative
        self.assertRaises(ValueError, s.run, -1)

        # Test running without a protocol
        s.set_protocol(None)
        s.run(1)

        # Test step size is > 0
        self.assertRaises(ValueError, s.set_step_size, 0)

        # Test negative log interval is ignored
        s.run(1, log_interval=-1)

    def test_progress_reporter(self):
        # Test running with a progress reporter.
        m, p, x = myokit.load(os.path.join(DIR_DATA, 'lr-1991.mmt'))

        # Test using a progress reporter
        with WarningCollector() as c:
            s = myokit.ICSimulation(m, p)
        with myokit.tools.capture() as c:
            s.run(110, progress=myokit.ProgressPrinter())
        c = c.text().splitlines()
        self.assertTrue(len(c) > 0)

        # Not a progress reporter
        self.assertRaisesRegex(
            ValueError, 'ProgressReporter', s.run, 5, progress=12)

        # Cancel from reporter
        self.assertRaises(
            myokit.SimulationCancelledError, s.run, 1,
            progress=CancellingReporter(0))

    def test_invalid_model(self):
        # Test running with an invalid model.
        m = myokit.Model()
        with WarningCollector() as c:
            self.assertRaises(
                myokit.MissingTimeVariableError, myokit.ICSimulation, m)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Tests the classes in `myokit.lib.common`.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from __future__ import absolute_import, division
from __future__ import print_function, unicode_literals

import os
import unittest
import numpy as np

import myokit

from myokit.tests import DIR_DATA, WarningCollector


class CommonTest(unittest.TestCase):
    """Tests lib.common"""

    def test_activation(self):
        # Test the activation experiment class.
        with WarningCollector():
            import myokit.lib.common as common

        # Load model
        m = os.path.join(DIR_DATA, 'lr-1991.mmt')
        m = myokit.load_model(m)
        # Create experiment
        c = m.get('ina')
        g = c.add_variable('g')
        g.set_rhs('m^3 * h * j')
        m.validate()
        a = common.Activation(m, 'ina.g')
        # Test
        a.times()
        a.traces()
        a.peaks(normalize=True)
        a.peaks(normalize=False)
        a.fit_boltzmann()
        a.convert_g2i()
        a.times()
        a.traces()
        a.peaks(normalize=True)
        a.peaks(normalize=False)
        a.fit_boltzmann()

    def test_inactivation(self):
        # Test the inactivation experiment class.
        with WarningCollector():
            import myokit.lib.common as common

        # Load model
        m = os.path.join(DIR_DATA, 'lr-1991.mmt')
        m = myokit.load_model(m)
        # Create experiment
        c = m.get('ina')
        g = c.add_variable('g')
        g.set_rhs('m^3 * h * j')
        m.validate()
        a = common.Inactivation(m, 'ina.g')
        # Test
        a.times()
        a.traces()
        a.peaks(normalize=True)
        a.peaks(normalize=False)
        a.fit_boltzmann()

    def test_recovery(self):
        # Test the recovery experiment class.
        with WarningCollector():
            import myokit.lib.common as common

        # Load model
        m = os.path.join(DIR_DATA, 'lr-1991.mmt')
        m = myokit.load_model(m)
        # Create experiment
        c = m.get('ina')
        g = c.add_variable('g')
        g.set_rhs('m^3 * h * j')
        m.validate()
        r = common.Recovery(m, 'ina.g')
        # Test
        n = 20
        r.set_pause_duration(0.1, 10, n)
        d = r.ratio()
        self.assertEqual(len(d), 2)
        self.assertIn('engine.time', d)
        self.assertIn('ina.g', d)
        self.assertEqual(n, len(d['engine.time']))
        self.assertEqual(n, len(d['ina.g']))
        x = d['engine.time']
        self.assertTrue(np.all(x[1:] > x[:-1]))
        x = d['ina.g']  # This is a monotonically increasing function
        self.assertTrue(np.all(x[1:] > x[:-1]))

    def test_restitution(self):
        # Test the restitution experiment class.
        with WarningCollector():
            import myokit.lib.common as common

        # Load model
        m = os.path.join(DIR_DATA, 'lr-1991.mmt')
        m = myokit.load_model(m)
        # Create experiment
        r = common.Restitution(m)
        # Test
        r.set_times(300, 800, 200)
        r.run()

    def test_strength_duration(self):
        # Test the strength-duration experiment class.
        with WarningCollector():
            import myokit.lib.common as common

        # Load model
        m = os.path.join(DIR_DATA, 'lr-1991.mmt')
        m = myokit.load_model(m)
        # Create experiment
        s = common.StrengthDuration(m, 'membrane.i_stim')
        # Test
        s.set_currents(-200, -10)
        s.set_times(0.5, 1.0, 0.2)
        s.run()


if __name__ == '__main__':
    unittest.main()



#
# Qt gui for viewing DataBlock2d data files.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from __future__ import absolute_import, division
from __future__ import print_function, unicode_literals

# Standard library imports
import gc
import os
import sys
import traceback

# Qt imports
from myokit.gui import QtWidgets, QtGui, QtCore, Qt

# Myokit
import myokit
import myokit.gui

# Myokit components
import myokit.formats.axon
import myokit.formats.wcp

# Matplotlib (must be imported _after_ gui has had chance to set backend)
import matplotlib
import matplotlib.figure
from myokit.gui import matplotlib_backend as backend

# NumPy
import numpy as np

# ConfigParser in Python 2 and 3
try:
    import ConfigParser as configparser
except ImportError:
    import configparser


# Application title
TITLE = 'Myokit DataLog Viewer (PROTOTYPE)'

# Application icon
# def icon():
#    icons = [
#        'icon-datalog-viewer.ico',
#        'icon-datalog-viewer-16.xpm',
#        'icon-datalog-viewer-24.xpm',
#        'icon-datalog-viewer-32.xpm',
#        'icon-datalog-viewer-48.xpm',
#        'icon-datalog-viewer-64.xpm',
#        'icon-datalog-viewer-96.xpm',
#        'icon-datalog-viewer-128.xpm',
#        'icon-datalog-viewer-256.xpm',
#        ]
#    icon = QtGui.QIcon()
#    for i in icons:
#        icon.addFile(os.path.join(myokit.DIR_DATA, 'gui', i))
#    return icon

# Settings file
SETTINGS_FILE = os.path.join(myokit.DIR_USER, 'DataLogViewer.ini')
# Number of recent files to display
# N_RECENT_FILES = 5

# About
ABOUT = '<h1>' + TITLE + '</h1>' + """
<p>
    The DataLog viewer is a PROTOTYPE utility to examine time series data.
    At the moment, exclusively WinWCP, ABF and CSV files.
</p>
<p>
    System info:
    <br />Python: PYTHON
    <br />Using the BACKEND GUI backend.
</p>
""".replace('BACKEND', myokit.gui.backend).replace('PYTHON', sys.version)

# License
LICENSE = myokit.LICENSE_HTML

# File filters
FILTER_ABF = 'ABF files (*.abf *.pro)'
FILTER_ATF = 'ATF files (*.atf)'
FILTER_CSV = 'CSV files (*.csv)'
FILTER_MAT = 'MAT files (*.mat)'
FILTER_TXT = 'TXT files (*.txt)'
FILTER_WCP = 'WCP files (*.wcp)'
FILTER_ZIP = 'Zipped DataLog files (*.zip)'
FILTER_ANY = 'All files (*.*)'
FILTER_ALL = 'Data files (*.abf *.csv *.mat *.pro *.txt *.wcp *.zip)'
FILTER_LIST = ';;'.join([
    FILTER_ALL,
    FILTER_ABF,
    FILTER_CSV,
    FILTER_MAT,
    FILTER_TXT,
    FILTER_WCP,
    FILTER_ZIP,
    FILTER_ANY,
])


class DataLogViewer(myokit.gui.MyokitApplication):
    """
    Graphical interface for viewing DataLog data.
    """
    def __init__(self, *filenames):
        super(DataLogViewer, self).__init__()
        # Set Title, icon
        self.setWindowTitle(TITLE + ' ' + myokit.__version__)
        # Set size, center
        self.resize(800, 600)
        qr = self.frameGeometry()
        cp = QtWidgets.QDesktopWidget().availableGeometry().center()
        qr.moveCenter(cp)
        self.move(qr.topLeft())
        # Add widget for Abf file tabs
        self._tabs = QtWidgets.QTabWidget()
        self._tabs.setTabsClosable(True)
        self._tabs.tabCloseRequested.connect(self.action_close)
        self.setCentralWidget(self._tabs)
        # Menu bar
        self.create_menu()
        # Tool bar
        self.create_toolbar()
        # Status bar
        self.statusBar().showMessage('Ready')
        # Current path
        self._path = QtCore.QDir.currentPath()
        # Load settings from ini file
        self.load_config()
        # Load any selected files
        for filename in filenames:
            self.load_file(filename)

    def action_about(self):
        """
        Displays the about dialog.
        """
        QtWidgets.QMessageBox.about(self, TITLE, ABOUT)

    def action_close(self, index):
        """
        Called when a tab should be closed
        """
        tab = self._tabs.widget(index)
        self._tabs.removeTab(index)
        if tab is not None:
            tab.deleteLater()
        gc.collect()
        del(tab)

    def action_license(self):
        """
        Displays this program's licensing information.
        """
        QtWidgets.QMessageBox.about(self, TITLE, LICENSE)

    def action_open(self):
        """
        Let the user select and open a file.
        """
        filenames = QtWidgets.QFileDialog.getOpenFileNames(
            self, 'Open data file', self._path, filter=FILTER_LIST)[0]
        if filenames:
            # Save current number of tabs
            tab_count = self._tabs.count()
            # Load files
            for filename in filenames:
                self.load_file(str(filename))
            # If loading went ok, show first of newly loaded files
            if self._tabs.count() > tab_count:
                self._tabs.setCurrentIndex(tab_count)

    def closeEvent(self, event=None):
        """
        Called when window is closed. To force a close (and trigger this
        function, call self.close())
        """
        # Save configuration
        self.save_config()
        if event:
            # Accept the event, close the window
            event.accept()
            # Ignore the event, window stays open
            #event.ignore()

    def create_menu(self):
        """
        Creates this widget's menu.
        """
        self._menu = self.menuBar()
        # File menu
        self._menu_file = self._menu.addMenu('&File')
        # File > Open
        self._tool_open = QtWidgets.QAction('&Open', self)
        self._tool_open.setShortcut('Ctrl+O')
        self._tool_open.setStatusTip('Open a file')
        self._tool_open.setIcon(QtGui.QIcon.fromTheme('document-open'))
        self._tool_open.triggered.connect(self.action_open)
        self._menu_file.addAction(self._tool_open)
        # File > ----
        self._menu_file.addSeparator()
        # File > Quit
        self._tool_exit = QtWidgets.QAction('&Quit', self)
        self._tool_exit.setShortcut('Ctrl+Q')
        self._tool_exit.setStatusTip('Exit application.')
        self._tool_exit.setIcon(QtGui.QIcon.fromTheme('application-exit'))
        self._tool_exit.triggered.connect(self.close)
        self._menu_file.addAction(self._tool_exit)
        # Help menu
        self._menu_help = self._menu.addMenu('&Help')
        # Help > About
        self._tool_about = QtWidgets.QAction('&About', self)
        self._tool_about.setStatusTip('View information about this program.')
        self._tool_about.triggered.connect(self.action_about)
        self._menu_help.addAction(self._tool_about)
        # Help > License
        self._tool_license = QtWidgets.QAction('&License', self)
        self._tool_license.setStatusTip('View this program\'s license info.')
        self._tool_license.triggered.connect(self.action_license)
        self._menu_help.addAction(self._tool_license)

    def create_toolbar(self):
        """
        Creates this widget's toolbar
        """
        self._toolbar = self.addToolBar('tools')
        self._toolbar.setFloatable(False)
        self._toolbar.setMovable(False)
        self._toolbar.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
        self._toolbar.addAction(self._tool_open)
        #self._toolbar.addSeparator()

    def load_config(self):
        """
        Loads the user configuration from an ini file.
        """
        # Read ini file
        config = configparser.RawConfigParser()
        try:
            config.read(os.path.expanduser(SETTINGS_FILE))
        except configparser.ParsingError:
            # Partially read config causes all sorts of errors, so discard
            config = configparser.RawConfigParser()

        # Window dimensions and location
        if config.has_section('window'):
            g = self.geometry()

            def getor(name, alt):
                if config.has_option('window', name):
                    return int(config.get('window', name))
                return alt

            x = getor('x', g.x())
            y = getor('y', g.y())
            w = getor('w', g.width())
            h = getor('h', g.height())
            self.setGeometry(x, y, w, h)

        # Current files, directory, etc
        if config.has_section('files'):
            if config.has_option('files', 'path'):
                path = config.get('files', 'path')
                if os.path.isdir(path):
                    self._path = path

    def load_file(self, filename):
        """
        Loads a data file.
        """
        root, ext = os.path.splitext(os.path.basename(filename))
        actions = {
            '.abf': self.load_abf_file,
            '.atf': self.load_atf_file,
            '.csv': self.load_datalog,
            '.mat': self.load_mat_file,
            '.pro': self.load_abf_file,
            '.txt': self.load_txt_file,
            '.wcp': self.load_wcp_file,
            '.zip': self.load_datalog,
        }
        try:
            action = actions[ext.lower()]
        except KeyError:
            QtWidgets.QMessageBox.critical(
                self, TITLE, 'File format not recognized: ' + ext)
            return
        action(filename)

    def load_abf_file(self, filename):
        """
        Loads an abf file.
        """
        try:
            abf = myokit.formats.axon.AbfFile(filename)
        except Exception:
            e = traceback.format_exc()
            QtWidgets.QMessageBox.critical(self, TITLE, e)
            return
        self._path = os.path.dirname(filename)
        self._tabs.addTab(AbfTab(self, abf), os.path.basename(filename))

    def load_atf_file(self, filename):
        """
        Loads an ATF file.
        """
        try:
            atf = myokit.formats.axon.AtfFile(filename)
        except Exception:
            e = traceback.format_exc()
            QtWidgets.QMessageBox.critical(self, TITLE, e)
            return
        self._path = os.path.dirname(filename)
        self._tabs.addTab(AtfTab(self, atf), os.path.basename(filename))

    def load_datalog(self, filename):
        """
        Loads a DataLog from csv or zip file.
        """
        try:
            if filename[-4:].lower() == '.csv':
                log = myokit.DataLog.load_csv(filename)
            else:
                log = myokit.DataLog.load(filename)
            if log.time_key is None:
                raise Exception('Log must contain a suitable time variable.')
        except Exception:
            e = traceback.format_exc()
            QtWidgets.QMessageBox.critical(self, TITLE, e)
            return
        self._path = os.path.dirname(filename)
        name = os.path.basename(filename)
        self._tabs.addTab(CsvTab(self, log, name), name)

    def load_mat_file(self, filename):
        """
        Loads a Matlab file.

        This method requires ``SciPy`` to be installed.
        """
        try:
            from scipy.io import loadmat
            mat = loadmat(filename)
        except Exception:
            e = traceback.format_exc()
            QtWidgets.QMessageBox.critical(self, TITLE, e)
            return
        self._path = os.path.dirname(filename)
        name = os.path.basename(filename)
        self._tabs.addTab(MatTab(self, mat, name), name)

    def load_txt_file(self, filename):
        """
        Loads a csv file.
        """
        try:
            data = np.loadtxt(filename)
        except Exception:
            e = traceback.format_exc()
            QtWidgets.QMessageBox.critical(self, TITLE, e)
            return
        self._path = os.path.dirname(filename)
        name = os.path.basename(filename)
        self._tabs.addTab(TxtTab(self, data, name), name)

    def load_wcp_file(self, filename):
        """
        Loads a wcp file.
        """
        try:
            wcp = myokit.formats.wcp.WcpFile(filename)
        except Exception:
            e = traceback.format_exc()
            QtWidgets.QMessageBox.critical(self, TITLE, e)
            return
        self._path = os.path.dirname(filename)
        self._tabs.addTab(WcpTab(self, wcp), os.path.basename(filename))

    def save_config(self):
        """
        Saves the user configuration to an ini file.
        """
        config = configparser.RawConfigParser()

        # Window dimensions and location
        config.add_section('window')
        g = self.geometry()
        config.set('window', 'x', str(g.x()))
        config.set('window', 'y', str(g.y()))
        config.set('window', 'w', str(g.width()))
        config.set('window', 'h', str(g.height()))

        # Current files, directory, etc
        config.add_section('files')
        config.set('files', 'path', self._path)

        # Write configuration to ini file
        inifile = os.path.expanduser(SETTINGS_FILE)
        with open(inifile, 'w') as configfile:
            config.write(configfile)

    def show(self):
        """
        Shows this viewer.
        """
        super(DataLogViewer, self).show()
        QtWidgets.QApplication.processEvents()


class AbfTab(QtWidgets.QTabWidget):
    """
    A widget displaying an ABF file.
    """
    def __init__(self, parent, abf):
        super(AbfTab, self).__init__(parent)
        self.setTabsClosable(False)
        self.setTabPosition(self.East)
        self._abf = abf
        self._figures = []
        self._axes = []
        for i in range(self._abf.data_channels()):
            tab, name = self.create_graph_tab(i)
            self.addTab(tab, name)
        for i in range(self._abf.protocol_channels()):
            tab, name = self.create_protocol_tab(i)
            self.addTab(tab, name)
        self.addTab(self.create_info_tab(), 'Info')
        del(self._abf)

    def create_graph_tab(self, channel):
        """
        Creates a widget displaying the main data.
        """
        widget = QtWidgets.QWidget(self)
        # Create figure
        figure = matplotlib.figure.Figure()
        figure.suptitle(self._abf.filename())
        canvas = backend.FigureCanvasQTAgg(figure)
        canvas.setParent(widget)
        axes = figure.add_subplot(1, 1, 1)
        toolbar = backend.NavigationToolbar2QT(canvas, widget)
        # Draw lines
        name = 'AD(' + str(channel) + ')'   # Default if no data is present
        times = None
        for i, sweep in enumerate(self._abf):
            if times is None:
                name = 'AD' + str(sweep[channel].number()) + ': ' \
                    + sweep[channel].name()
                times = sweep[channel].times()
            axes.plot(times, sweep[channel].values())
        # Create a layout
        vbox = QtWidgets.QVBoxLayout()
        vbox.addWidget(canvas)
        vbox.addWidget(toolbar)
        widget.setLayout(vbox)
        self._figures.append(figure)
        self._axes.append(axes)
        return widget, name

    def create_protocol_tab(self, channel):
        """
        Creates a widget displaying a stored D/A signal.
        """
        widget = QtWidgets.QWidget(self)
        # Create figure
        figure = matplotlib.figure.Figure()
        figure.suptitle(self._abf.filename())
        canvas = backend.FigureCanvasQTAgg(figure)
        canvas.setParent(widget)
        axes = figure.add_subplot(1, 1, 1)
        toolbar = backend.NavigationToolbar2QT(canvas, widget)
        # Draw lines
        name = 'DA(' + str(channel) + ')'   # Default if no data is present
        times = None
        for i, sweep in enumerate(self._abf.protocol()):
            if times is None:
                name = 'DA' + str(sweep[channel].number()) + ': ' \
                    + sweep[channel].name()
                times = sweep[channel].times()
            axes.plot(times, sweep[channel].values())
        # Create a layout
        vbox = QtWidgets.QVBoxLayout()
        vbox.addWidget(canvas)
        vbox.addWidget(toolbar)
        widget.setLayout(vbox)
        self._figures.append(figure)
        self._axes.append(axes)
        return widget, name

    def create_info_tab(self):
        """
        Creates a tab displaying information about the file.
        """
        widget = QtWidgets.QTextEdit(self)
        widget.setText(self._abf.info(show_header=True))
        widget.setReadOnly(True)
        return widget

    def deleteLater(self):
        """
        Deletes this tab (later).
        """
        for figure in self._figures:
            figure.clear()
        for axes in self._axes:
            axes.cla()
        del(self._figures, self._axes)
        gc.collect()
        super(AbfTab, self).deleteLater()


class AtfTab(QtWidgets.QTabWidget):
    """
    A widget displaying an AGF file.
    """
    def __init__(self, parent, atf):
        super(AtfTab, self).__init__(parent)
        self._atf = atf

        self.setTabsClosable(False)
        self.setTabPosition(self.East)

        self._figures = []
        self._axes = []
        keys = list(self._atf.keys())
        if len(keys) > 1:
            time = keys[0]  # Time is always first (and regularly sampled)
            for key in keys[1:]:
                self.addTab(self.create_graph_tab(time, key), key)
        self.addTab(self.create_info_tab(), 'Info')
        del(self._atf)

    def create_graph_tab(self, time, key):
        """
        Creates a widget displaying a graph.
        """
        widget = QtWidgets.QWidget(self)

        # Create figure
        figure = matplotlib.figure.Figure()
        figure.suptitle(self._atf.filename())
        canvas = backend.FigureCanvasQTAgg(figure)
        canvas.setParent(widget)
        axes = figure.add_subplot(1, 1, 1)
        toolbar = backend.NavigationToolbar2QT(canvas, widget)

        # Draw lines
        axes.plot(self._atf[time], self._atf[key])

        # Create a layout
        vbox = QtWidgets.QVBoxLayout()
        vbox.addWidget(canvas)
        vbox.addWidget(toolbar)
        widget.setLayout(vbox)
        self._figures.append(figure)
        self._axes.append(axes)

        # Return widget
        return widget

    def create_info_tab(self):
        """
        Creates a tab displaying information about the file.
        """
        widget = QtWidgets.QTextEdit(self)
        widget.setText(self._atf.info())
        widget.setReadOnly(True)
        return widget

    def deleteLater(self):
        """
        Deletes this tab (later).
        """
        for figure in self._figures:
            figure.clear()
        for axes in self._axes:
            axes.cla()
        del(self._figures, self._axes)
        gc.collect()
        super(AtfTab, self).deleteLater()


class CsvTab(QtWidgets.QTabWidget):
    """
    A widget displaying a CSV file.

    The given log must have a time variable set.
    """
    def __init__(self, parent, log, filename):
        super(CsvTab, self).__init__(parent)
        self.setTabsClosable(False)
        self.setTabPosition(self.East)
        self._log = log.npview()
        self._filename = filename
        self._figures = []
        self._axes = []

        # Check time key was found
        time = log.time_key()
        try:
            self._time = log.time()
        except myokit.InvalidDataLogError:
            if time is None:
                QtWidgets.QMessageBox.critical(
                    self, TITLE,
                    'Unable to load file: no time key set in this log.')
                return
            else:
                raise

        # Check that time series aren't empty
        if log.length() == 0:
            QtWidgets.QMessageBox.critical(
                self, TITLE, 'Unable to load file: no data found.')
            return

        # Add tab for each column
        for k, v in log.items():
            if k == time:
                continue
            self.addTab(self.create_graph_tab(k, v), k)

    def create_graph_tab(self, key, data):
        """
        Creates a widget displaying the ``data`` stored under ``key``.
        """
        widget = QtWidgets.QWidget(self)
        # Create figure
        figure = matplotlib.figure.Figure()
        figure.suptitle(self._filename)
        canvas = backend.FigureCanvasQTAgg(figure)
        canvas.setParent(widget)
        axes = figure.add_subplot(1, 1, 1)
        axes.set_title(key)
        toolbar = backend.NavigationToolbar2QT(canvas, widget)
        # Draw lines
        axes.plot(self._time, data)
        # Create a layout
        vbox = QtWidgets.QVBoxLayout()
        vbox.addWidget(canvas)
        vbox.addWidget(toolbar)
        widget.setLayout(vbox)
        self._figures.append(figure)
        self._axes.append(axes)
        return widget

    def deleteLater(self):
        """
        Deletes this tab (later).
        """
        for figure in self._figures:
            figure.clear()
        for axes in self._axes:
            axes.cla()
        del(self._figures, self._axes)
        gc.collect()
        super(CsvTab, self).deleteLater()


class MatTab(QtWidgets.QTabWidget):
    """
    A widget displaying a MAT file.
    """
    def __init__(self, parent, mat, filename):
        super(MatTab, self).__init__(parent)
        self.setTabsClosable(False)
        self.setTabPosition(self.East)
        self._figures = []
        self._filename = filename
        self._axes = []

        # Find usable data
        for key in mat.keys():
            if key[:1] == '_':
                continue
            time, data = None, None
            data = mat[key]
            if np.prod(data.shape) == np.max(data.shape):
                # 1d data
                time = None
                data = data.reshape((np.max(data.shape),))
            elif len(data.shape) == 2 and (
                    np.prod(data.shape) == 2 * np.max(data.shape)):
                # 2d data: Only allow len(shape) == 2, otherwise too many cases
                if data.shape[0] == 2:
                    time = data[0]
                    data = data[1]
                else:
                    time = data[:, 0]
                    data = data[:, 1]
                # Check time is increasing
                if np.any(time[1:] < time[:-1]):
                    time, data = data, time
                if np.any(time[1:] < time[:-1]):
                    time, data = None, None
            if data is None:
                continue
            # Create tab
            tab = self.create_graph_tab(time, data)
            self.addTab(tab, key)

        # Nothing that can be used? Show error
        if len(self._figures) == 0:
            QtWidgets.QMessageBox.critical(
                self, TITLE,
                'Unable to load file: no usable data found.')
            return

    def create_graph_tab(self, time, data):
        """
        Creates a widget displaying a time series.
        """
        widget = QtWidgets.QWidget(self)
        # Create figure
        figure = matplotlib.figure.Figure()
        figure.suptitle(self._filename)
        canvas = backend.FigureCanvasQTAgg(figure)
        canvas.setParent(widget)
        axes = figure.add_subplot(1, 1, 1)
        toolbar = backend.NavigationToolbar2QT(canvas, widget)
        # Draw line
        if time is None:
            axes.plot(data)
        else:
            axes.plot(time, data)
        # Create a layout
        vbox = QtWidgets.QVBoxLayout()
        vbox.addWidget(canvas)
        vbox.addWidget(toolbar)
        widget.setLayout(vbox)
        self._figures.append(figure)
        self._axes.append(axes)
        return widget

    def deleteLater(self):
        """
        Deletes this tab (later).
        """
        for figure in self._figures:
            figure.clear()
        for axes in self._axes:
            axes.cla()
        del(self._figures, self._axes)
        gc.collect()
        super(MatTab, self).deleteLater()


class TxtTab(QtWidgets.QTabWidget):
    """
    A widget displaying a TXT file (with lots of heuristics!).
    """
    def __init__(self, parent, data, filename):
        super(TxtTab, self).__init__(parent)
        self.setTabsClosable(False)
        self.setTabPosition(self.East)
        self._figures = []
        self._filename = filename
        self._axes = []

        # Find usable data
        if np.prod(data.shape) == np.max(data.shape):
            # 1d data
            data = data.reshape((np.max(data.shape),))
            # Create tab
            tab = self.create_graph_tab(None, data)
            self.addTab(tab, 'series 1')
        elif len(data.shape) == 2:
            # 2d data, assume longest axis is time
            if data.shape[0] < data.shape[1]:
                data = data.T
            # Check if first or last entry could be time
            if not np.any(data[0, 1:] < data[0, :-1]):
                time = data[0]
                data = data[1:]
            elif not np.any(data[-1, 1:] < data[-1, :-1]):
                time = data[-1]
                data = data[:-1]
            else:
                time = None
            # Create tabs
            for k, column in enumerate(data):
                tab = self.create_graph_tab(time, data)
                self.addTab(tab, 'series ' + str(1 + k))
        else:
            # Nothing that can be used? Show error
            QtWidgets.QMessageBox.critical(
                self, TITLE,
                'Unable to load file: unable to parse file contents.')
            return

    def create_graph_tab(self, time, data):
        """
        Creates a widget displaying a time series.
        """
        widget = QtWidgets.QWidget(self)
        # Create figure
        figure = matplotlib.figure.Figure()
        figure.suptitle(self._filename)
        canvas = backend.FigureCanvasQTAgg(figure)
        canvas.setParent(widget)
        axes = figure.add_subplot(1, 1, 1)
        toolbar = backend.NavigationToolbar2QT(canvas, widget)
        # Draw line
        if time is None:
            axes.plot(data)
        else:
            axes.plot(time, data)
        # Create a layout
        vbox = QtWidgets.QVBoxLayout()
        vbox.addWidget(canvas)
        vbox.addWidget(toolbar)
        widget.setLayout(vbox)
        self._figures.append(figure)
        self._axes.append(axes)
        return widget

    def deleteLater(self):
        """
        Deletes this tab (later).
        """
        for figure in self._figures:
            figure.clear()
        for axes in self._axes:
            axes.cla()
        del(self._figures, self._axes)
        gc.collect()
        super(TxtTab, self).deleteLater()


class WcpTab(QtWidgets.QTabWidget):
    """
    A widget displaying a WCP file.
    """
    def __init__(self, parent, wcp):
        super(WcpTab, self).__init__(parent)
        self.setTabsClosable(False)
        self.setTabPosition(self.East)
        self._wcp = wcp
        self._figures = []
        self._axes = []
        for i in range(self._wcp.records()):
            self.addTab(self.create_graph_tab(i), 'Record ' + str(i))
        del(self._wcp)

    def create_graph_tab(self, record):
        """
        Creates a widget displaying the data in record i
        """
        widget = QtWidgets.QWidget(self)
        # Create figure
        figure = matplotlib.figure.Figure()
        figure.suptitle(self._wcp.filename())
        canvas = backend.FigureCanvasQTAgg(figure)
        canvas.setParent(widget)
        axes = figure.add_subplot(1, 1, 1)
        toolbar = backend.NavigationToolbar2QT(canvas, widget)
        # Draw lines
        for i in range(self._wcp.channels()):
            axes.plot(
                np.array(self._wcp.times(), copy=True),
                np.array(self._wcp.values(record, i), copy=True),
            )
        # Create a layout
        vbox = QtWidgets.QVBoxLayout()
        vbox.addWidget(canvas)
        vbox.addWidget(toolbar)
        widget.setLayout(vbox)
        self._figures.append(figure)
        self._axes.append(axes)
        return widget

    def deleteLater(self):
        """
        Deletes this tab (later).
        """
        for figure in self._figures:
            figure.clear()
        for axes in self._axes:
            axes.cla()
        del(self._figures, self._axes)
        gc.collect()
        super(WcpTab, self).deleteLater()




#!/usr/bin/env python3
#
# Tests the Axon format module.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from __future__ import absolute_import, division
from __future__ import print_function, unicode_literals

import os
import unittest
import numpy as np

import myokit
import myokit.formats.axon as axon

from myokit.tests import TemporaryDirectory, DIR_FORMATS

# Unit testing in Python 2 and 3
try:
    unittest.TestCase.assertRaisesRegex
except AttributeError:
    unittest.TestCase.assertRaisesRegex = unittest.TestCase.assertRaisesRegexp


class AbfTest(unittest.TestCase):
    """
    Tests the ABF format support.
    """

    def test_read_v1(self):
        # Test reading a version 1 file.

        # Load file
        path = os.path.join(DIR_FORMATS, 'abf-v1.abf')
        abf = axon.AbfFile(path)
        self.assertEqual(abf.filename(), path)

        # Check version info
        self.assertIn('version 1.65', abf.info())
        self.assertEqual(
            abf.info(),
            'Axon Binary File: abf-v1.abf\n'
            'ABF Format version 1.65\n'
            'Recorded on: 2014-11-14 12:52:29.389999\n'
            'Acquisition mode: 5: Episodic stimulation mode\n'
            'Protocol set for 1 trials, spaced 0.0s apart.\n'
            '    with 1 runs per trial, spaced 0.0s apart.\n'
            '     and 9 sweeps per run, spaced 0.5s apart.\n'
            'Sampling rate: 10000.0 Hz\n'
            'Channel 0: "IN 0      "\n'
            '  Unit: pA'
        )
        # Test getting full header runs without crashing
        abf.info(True)

        # Test len returns number of sweeps
        self.assertEqual(len(abf), 9)

        # Test data access
        self.assertEqual(abf.data_channels(), 1)    # 1 data channel
        x = abf.extract_channel(0)
        self.assertEqual(len(x), 1 + len(abf))      # sweeps + time
        self.assertEqual(len(x[0]), len(x[1]))
        self.assertEqual(len(x[0]), len(x[2]))
        self.assertEqual(len(x[0]), len(x[3]))
        self.assertEqual(len(x[0]), len(x[4]))
        self.assertEqual(len(x[0]), len(x[5]))
        self.assertEqual(len(x[0]), len(x[6]))
        self.assertEqual(len(x[0]), len(x[7]))
        self.assertEqual(len(x[0]), len(x[8]))
        self.assertEqual(len(x[0]), len(x[9]))
        y = abf.extract_channel_as_myokit_log(0)
        self.assertEqual(len(y), 1 + len(abf))      # sweeps + time
        z = abf.myokit_log()
        self.assertEqual(len(z), 6)     # time + channel + 4 protocol channels
        sweep = abf[0]
        self.assertEqual(len(sweep), 1)     # 1 channel in sweep
        channel = sweep[0]
        self.assertIsInstance(channel.number(), int)
        self.assertIn('Channel', str(channel))
        self.assertEqual(len(abf) * len(channel.times()), len(z.time()))
        self.assertEqual(len(abf) * len(channel.values()), len(z.time()))

        # Test reading of sweeps as one long array
        x, y = abf.extract_channel(0, join=True)
        z = abf.extract_channel(0)
        self.assertEqual(len(x), len(y))
        self.assertEqual(len(x), len(abf) * len(z[0]))
        self.assertTrue(np.all(x[1:] > x[:-1]))

        # Test protocol extraction
        self.assertEqual(abf.protocol_channels(), 4)    # 4 protocol channels
        p = abf.myokit_protocol()
        self.assertEqual(len(p), 18)    # 18 steps in this protocol
        self.assertEqual(abf.protocol_holding_level(0), 0)
        p = abf.myokit_protocol(0)
        self.assertEqual(len(p), 18)

    def test_read_v2(self):
        # Test reading a version 2 file.

        # Load file
        path = os.path.join(DIR_FORMATS, 'abf-v2.abf')
        abf = axon.AbfFile(path)

        # Check version info
        self.assertIn('version 2.0', abf.info())
        self.assertEqual(
            abf.info(),
            'Axon Binary File: abf-v2.abf\n'
            'ABF Format version 2.0\n'
            'Recorded on: 2014-10-01 14:03:55.980999\n'
            'Acquisition mode: 5: Episodic stimulation mode\n'
            'Protocol set for 1 trials, spaced 0.0s apart.\n'
            '    with 1 runs per trial, spaced 0.0s apart.\n'
            '     and 1 sweeps per run, spaced 5.0s apart.\n'
            'Sampling rate: 10000.0 Hz\n'
            'Channel 0: "IN 0"\n'
            '  Unit: pA\n'
            '  Low-pass filter: 10000.0 Hz\n'
            '  Cm (telegraphed): 6.34765625 pF')
        # Test getting full header runs without crashing
        abf.info(True)

        # Test len returns number of sweeps
        self.assertEqual(len(abf), 1)

        # Test data access
        self.assertEqual(abf.data_channels(), 1)    # 1 data channel
        x = abf.extract_channel(0)
        self.assertEqual(len(x), 1 + len(abf))      # sweeps + time
        self.assertEqual(len(x[0]), len(x[1]))
        y = abf.extract_channel_as_myokit_log(0)
        self.assertEqual(len(y), 1 + len(abf))      # sweeps + time
        z = abf.myokit_log()
        self.assertEqual(len(z), 6)     # time + channel + 4 protocol channels
        sweep = abf[0]
        self.assertEqual(len(sweep), 1)     # 1 channel in sweep
        channel = sweep[0]
        self.assertEqual(len(abf) * len(channel.times()), len(z.time()))
        self.assertEqual(len(abf) * len(channel.values()), len(z.time()))

        # Test protocol extraction
        self.assertEqual(abf.protocol_channels(), 4)    # 4 protocol channels
        p = abf.myokit_protocol()
        self.assertEqual(len(p), 2)     # 2 steps in this protocol
        self.assertEqual(abf.protocol_holding_level(0), -120)
        p = abf.myokit_protocol(0)
        self.assertEqual(len(p), 2)

    def test_read_protocol_v1(self):
        # Test reading a v1 protocol file.

        # Load file
        path = os.path.join(DIR_FORMATS, 'abf-protocol.pro')
        abf = axon.AbfFile(path)

        # Check version info
        self.assertIn('version 1.65', abf.info())
        self.assertEqual(
            abf.info(),
            'Axon Protocol File: abf-protocol.pro\n'
            'ABF Format version 1.65\n'
            'Recorded on: 2005-06-17 14:33:02.160000\n'
            'Acquisition mode: 5: Episodic stimulation mode\n'
            'Protocol set for 1 trials, spaced 0.0s apart.\n'
            '    with 1 runs per trial, spaced 0.0s apart.\n'
            '     and 30 sweeps per run, spaced 5.0s apart.\n'
            'Sampling rate: 20000.0 Hz'
        )
        # Test getting full header runs without crashing
        abf.info(True)

        # Load, force as protocol
        path = os.path.join(DIR_FORMATS, 'abf-protocol.pro')
        abf = axon.AbfFile(path, is_protocol_file=True)

        # Check version info
        self.assertIn('version 1.65', abf.info())
        self.assertIn('Axon Protocol File', abf.info())

        # Test protocol extraction
        p = abf.myokit_protocol()
        self.assertEqual(len(p), 60)

        # Test step extraction
        p = abf.protocol_steps()
        self.assertEqual(len(p), 1)
        self.assertEqual(len(p[0]), 30)

    def test_matplotlib_figure(self):
        # Test figure drawing method (doesn't inspect output).
        # Select matplotlib backend that doesn't require a screen
        import matplotlib
        matplotlib.use('template')
        path = os.path.join(DIR_FORMATS, 'abf-v1.abf')
        abf = axon.AbfFile(path)
        abf.matplotlib_figure()


class AtfTest(unittest.TestCase):
    """
    Tests the ATF format support.
    """

    def test_write_read(self):
        # Test writing and reading an ATF file.

        with TemporaryDirectory() as d:
            # Create data log
            log = myokit.DataLog()
            log.set_time_key('time')
            log['time'] = np.arange(100)
            log['sint'] = np.sin(log['time'])
            log['cost'] = np.cos(log['time'])

            # Write atf file
            path = d.path('test.atf')
            axon.save_atf(log, path)

            # Read atf file
            log2 = axon.load_atf(path)
            self.assertEqual(len(log), len(log2))
            self.assertEqual(set(log.keys()), set(log2.keys()))
            for k, v in log.items():
                self.assertTrue(np.all(v == log2[k]))

            # Write selected fields
            axon.save_atf(log, path, fields=['time', 'sint'])
            log2 = axon.load_atf(path)
            self.assertEqual(set(log2.keys()), set(['time', 'sint']))

            # Time must be regularly spaced
            log['time'][-1] *= 2
            self.assertRaisesRegex(
                ValueError, 'regularly spaced', axon.save_atf, log, path)

            # Field names can't contain quotes
            log['time'] = np.arange(100)
            log['si"nt'] = log['sint']
            self.assertRaisesRegex(
                ValueError, 'double quotes', axon.save_atf, log, path)

            # Field names can't have newlines
            del(log['si"nt'])
            log['si\nnt'] = log['sint']
            self.assertRaisesRegex(
                ValueError, 'newlines', axon.save_atf, log, path)

            # Fields in `fields` must exist
            del(log['si\nnt'])
            self.assertRaisesRegex(
                ValueError, 'not found', axon.save_atf, log, path,
                fields=['time', 'sint', 'hi'])

            # Try using on other formats
            log.save_csv(path)
            self.assertRaisesRegex(Exception, 'file type', axon.load_atf, path)

            # Try reading raw meta data (no key-value pairs)
            with open(path, 'w') as f:
                f.write('ATF\t1.0\n')
                f.write('1\t3\n')
                f.write('"Hello! This is raw meta data"\n')
                f.write('"time"\t"sint"\t"cost"\n')
                f.write('0\t0.0\t1.0\n')
                f.write('1\t10\t20\n')
                f.write('2\t30\t40\n')
            log2 = axon.load_atf(path)

            # Test invalid header detection
            with open(path, 'w') as f:
                f.write('ATF\t1.0\n')
                f.write('1\t3\n')
                f.write('Hello! This is raw meta data\n')
                f.write('"time"\t"sint"\t"cost"\n')
                f.write('0\t0.0\t1.0\n')
                f.write('1\t10\t20\n')
                f.write('2\t30\t40\n')
            self.assertRaisesRegex(
                Exception, 'double quotation', axon.load_atf, path)

            # Bad column headers
            with open(path, 'w') as f:
                f.write('ATF\t1.0\n')
                f.write('1\t3\n')
                f.write('"Hello! This is raw meta data"\n')
                f.write('Bonjou\t"time"\t"sint"\t"cost"\n')
                f.write('0\t0.0\t1.0\n')
                f.write('1\t10\t20\n')
                f.write('2\t30\t40\n')
            self.assertRaisesRegex(
                Exception, 'column headers', axon.load_atf, path)

            # Bad column headers
            with open(path, 'w') as f:
                f.write('ATF\t1.0\n')
                f.write('1\t3\n')
                f.write('"Hello! This is raw meta data"\n')
                f.write('time"\t"sint"\t"cost"\n')
                f.write('0\t0.0\t1.0\n')
                f.write('1\t10\t20\n')
                f.write('2\t30\t40\n')
            self.assertRaisesRegex(
                Exception, 'column headers', axon.load_atf, path)

            # Too many headers
            with open(path, 'w') as f:
                f.write('ATF\t1.0\n')
                f.write('1\t3\n')
                f.write('"Hello! This is raw meta data"\n')
                f.write('"Bonjour"\t"time"\t"sint"\t"cost"\n')
                f.write('0\t0.0\t1.0\n')
                f.write('1\t10\t20\n')
                f.write('2\t30\t40\n')
            self.assertRaisesRegex(
                Exception, 'found 4', axon.load_atf, path)

            # Commas as delimiter are ok
            with open(path, 'w') as f:
                f.write('ATF\t1.0\n')
                f.write('1\t3\n')
                f.write('"Hello! This is raw meta data"\n')
                f.write('"time","sint","cost"\n')
                f.write('0,0.0,1.0\n')
                f.write('1,10,20\n')
                f.write('2,30,40\n')
            axon.load_atf(path)

            # But can't mix them
            with open(path, 'w') as f:
                f.write('ATF\t1.0\n')
                f.write('1\t3\n')
                f.write('"Hello! This is raw meta data"\n')
                f.write('"time"\t"sint","cost"\n')
                f.write('0,0.0,1.0\n')
                f.write('1,10,20\n')
                f.write('2,30,40\n')
            self.assertRaisesRegex(
                Exception, 'Mixed delimiters', axon.load_atf, path)

            # Too many columns
            with open(path, 'w') as f:
                f.write('ATF\t1.0\n')
                f.write('1\t3\n')
                f.write('"Hello! This is raw meta data"\n')
                f.write('"time"\t"sint"\t"cost"\n')
                f.write('0\t0.0\t1.0\n')
                f.write('1\t10\t20\t100\n')
                f.write('2\t30\t40\n')
            self.assertRaisesRegex(
                Exception, 'Invalid data', axon.load_atf, path)

    def test_accessors(self):
        # Test various accessor methods of :class:`AtfFile`.

        with TemporaryDirectory() as d:
            # Create data log
            log = myokit.DataLog()
            log.set_time_key('time')
            log['time'] = np.arange(10)
            log['sint'] = np.sin(log['time'])
            log['cost'] = np.cos(log['time'])

            # Write atf file
            path = d.path('test.atf')
            axon.save_atf(log, path)

            # Read atf file
            atf = myokit.formats.axon.AtfFile(path)

            # Test filename()
            self.assertEqual(atf.filename(), path)

            # Test iter and getitem
            self.assertEqual(len(list(iter(atf))), 3)
            self.assertTrue(np.all(atf['time'] == log['time']))
            self.assertTrue(np.all(atf['sint'] == log['sint']))
            self.assertTrue(np.all(atf['cost'] == log['cost']))

            # Test items()
            items = list(atf.items())
            self.assertEqual(items[0][0], 'time')
            self.assertEqual(items[1][0], 'sint')
            self.assertEqual(items[2][0], 'cost')
            self.assertTrue(np.all(items[0][1] == log['time']))
            self.assertTrue(np.all(items[1][1] == log['sint']))
            self.assertTrue(np.all(items[2][1] == log['cost']))

            # Test keys()
            self.assertEqual(list(atf.keys()), ['time', 'sint', 'cost'])

            # Test values()
            values = list(atf.values())
            self.assertTrue(np.all(values[0] == log['time']))
            self.assertTrue(np.all(values[1] == log['sint']))
            self.assertTrue(np.all(values[2] == log['cost']))

            # Test len
            self.assertEqual(len(atf), 3)

            # Test info
            self.assertIn('myokit', atf.info())

            # Test version
            self.assertEqual(atf.version(), '1.0')


if __name__ == '__main__':
    unittest.main()
